"""This module contains the various math and system dynamics object
components necessary to make up a model and construct its equations.
"""

from collections.abc import Callable

import matplotlib.pyplot as plt
import numpy as np
import pytensor.tensor as pt
from IPython.display import Markdown, display
from pytensor.ifelse import ifelse

import reno
from reno.utils import ensure_scalar, is_static, latex_name, range_eq_latex

# ==================================================
# EQUATION BASE CLASS
# ==================================================
# The root of all math things, the tree of life :)


class EquationPart:
    """The base object that represents some portion/subtree of a compute
    or equation tree.

    The purpose of EquationParts is to allow constructing an equation as a
    "deferred computation" - essentially symbolically creating a compute
    tree of objects all with ``.eval()`` functions that get recursively run
    later on when a simulation is running/equation is computing (spawned with
    root ``.eval()`.)

    Note:
        In order to get nice semantics when constructing equations, all math
        operations applied to an EquationPart get replaced with their
        corresponding operation EquationPart. This means a ``5 + Scalar(4)``
        gets converted into an ``Operation("+", Scalar(5), Scalar(4))``

    Args:
        sub_equation_parts (List[EquationPart]): The list of all **immediate**
            child equation subtrees, e.g. for the operation ``a + b``, the
            root equation part is the operation ``+``, and ``[a, b]`` would be
            its sub parts. These are used to help search the tree/connect
            parts of diagrams up. TODO: probably not strictly defined/enforced
            enough within the code, write tests for this.
    """

    def __init__(self, sub_equation_parts: list["EquationPart"] = None):
        if sub_equation_parts is None:
            sub_equation_parts = []
        # allow raw numbers to be passed in to any constructor, and ensure
        # they're wrapped in Scalars if relevant. This makes it so
        # reno.ops.add(5, Scalar(4)) correctly becomes Scalar(5) + Scalar(4)
        for i, part in enumerate(sub_equation_parts):
            sub_equation_parts[i] = ensure_scalar(part)
        self.value = None
        self.sub_equation_parts = sub_equation_parts

    # ---- MATH OPERATION OVERLOADING/EQUATION PART REPLACEMENT ----

    def __add__(self, obj):
        return reno.ops.add(self, obj)

    def __sub__(self, obj):
        return reno.ops.sub(self, obj)

    def __neg__(self):
        return reno.ops.sub(0, self)

    def __mul__(self, obj):
        return reno.ops.mul(self, obj)

    def __truediv__(self, obj):
        return reno.ops.div(self, obj)

    def __mod__(self, obj):
        return reno.ops.mod(self, obj)

    def __radd__(self, obj):
        return reno.ops.add(obj, self)

    def __rsub__(self, obj):
        return reno.ops.sub(obj, self)

    def __rmul__(self, obj):
        return reno.ops.mul(obj, self)

    def __rtruediv__(self, obj):
        return reno.ops.div(obj, self)

    def __rmod__(self, obj):
        return reno.ops.mod(self, obj)

    def __lt__(self, obj):
        return reno.ops.lt(self, obj)

    def __le__(self, obj):
        return reno.ops.lte(self, obj)

    def __gt__(self, obj):
        return reno.ops.gt(self, obj)

    def __ge__(self, obj):
        return reno.ops.gte(self, obj)

    def __and__(self, obj):
        return reno.ops.bool_and(self, obj)

    def __rand__(self, obj):
        return reno.ops.bool_and(self, obj)

    def __or__(self, obj):
        return reno.ops.bool_or(self, obj)

    def __ror__(self, obj):
        return reno.ops.bool_or(self, obj)

    def __getitem__(self, obj):
        if isinstance(obj, slice):
            return reno.ops.slice(self, obj.start, obj.stop)
        return reno.ops.index(self, obj)

    # NOTE: yeaaah weird things break if we override the eq portion.
    # Leaving this here as a reminder to not try to do this.
    # use the equal not_equal below.
    # def __eq__(self, obj):
    #     if isinstance(obj, (int, float)):
    #         return reno.ops.eq(self, Scalar(obj))
    #     return reno.ops.eq(self, obj)
    #
    # def __ne__(self, obj):
    #     if isinstance(obj, (int, float)):
    #         return reno.ops.ne(self, Scalar(obj))
    #     return reno.ops.ne(self, obj)

    # fill in remaining arithmetic and boolean operators as needed,
    # see https://www.pythonmorsels.com/every-dunder-method/

    # ---- /MATH OPERATION OVERLOADING/EQUATION PART REPLACEMENT ----

    # ---- CLEANER MATH API ----
    # (similar to how numpy allows either np.max(a) or a.max())

    def series_max(self):
        # See note in ops.py, not using 'max' because of TrackedReference
        return reno.ops.series_max(self)

    def series_min(self):
        # See note in ops.py, not using 'min' because of TrackedReference
        return reno.ops.series_min(self)

    def sum(self):
        return reno.ops.sum(self)

    def equal(self, obj):
        return reno.ops.eq(self, obj)

    def not_equal(self, obj):
        return reno.ops.ne(self, obj)

    def clip(self, min, max):
        return reno.ops.clip(self, min, max)

    # ---- /CLEANER MATH API ----

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """Execute the compute graph for this equation, this needs to be
        implemented in every subclass.

        Note that throughout a compute tree, this should effectively recurse
        through ``.eval()`` calls to all subparts as well.

        Args:
            t (int): Timestep along simulation at which to evaluate.
            save (bool): Whether to store/track/cache the output in a tracked
                matrix. This is really only applicable to ``TrackedReference``s,
                but given recursive nature of this function, needs to always be
                passed down through all subsequent calls.
            row_indices (list[int] | np.ndarray): The list of rows to evaluate
                and return. Similar to save, this is mostly only applicable to
                references that are actual matrices TODO: expand, mention piecewise.
                None means to evaluate all rows
            force (bool): Whether to ignore a previously cached value and compute
                regardless.
        """
        raise NotImplementedError()

    def latex(self, **kwargs) -> str:
        """Construct a string representation of this portion of the equation
        for use in a latex display. Should probably be overriden in most
        subclasses, and often needs to be called recursively on
        sub_equation_parts."""
        return str(self.value)

    def seek_refs(self):
        """Immediate refs only, depth=1."""
        # TODO: really not sure this is the best way of doing this, and it
        # heavily impacts quality of graph viz rendering/determining how
        # flows connect to what stocks and eachother etc. (can be implicit
        # connections if equations are set up weird.)
        refs = []
        check_parts = [*self.sub_equation_parts]
        if isinstance(self.value, EquationPart):
            # TODO: I think this handles cases where you've directly assigned a
            # scalar as a value to something?
            check_parts.append(self.value)
        if isinstance(self, TrackedReference):
            if self.min is not None:
                check_parts.append(self.min)
            if self.max is not None:
                check_parts.append(self.max)
        for part in check_parts:
            if isinstance(part, Function):
                refs.extend(part.seek_refs())
            elif isinstance(part, Operation):
                refs.extend(part.seek_refs())
            elif isinstance(part, Piecewise):
                refs.extend(part.seek_refs())

            elif isinstance(part, HistoricalValue):
                refs.append(part.tracked_ref)
                refs.append(part)  # TODO: (2025.02.03) ?????
                # I feel like I'm still missing something between eq setting, assign op, seek_refs and static checks, but test_static_check_eq_with_historical_value breaks without this
                refs.extend(part.index_eq.seek_refs())
            elif isinstance(part, Reference):
                refs.append(part)

            # make sure anything with limits counts stuff in those?
            # (not sure if this is actually necessary?)
            # if isinstance(part, Variable) or isinstance(part, Flow):
            #     if part.min is not None:
            #         print("Adding min refs")
            #         refs.extend(part.min.seek_refs())
            #     if part.max is not None:
            #         print("Adding min refs")
            #         refs.extend(part.max.seek_refs())
        # be sure to remove duplicates
        return list(set(refs))

    def find_refs_of_type(self, search_type, already_checked: list = None) -> list:
        """Actually recursive as opposed to seek_refs, returns a list of all equation
        parts matching passed type."""

        if already_checked is None:
            already_checked = []

        refs = []
        check_parts = [*self.sub_equation_parts]
        if isinstance(self.value, EquationPart):
            # TODO: I think this handles cases where you've directly assigned a
            # scalar as a value to something?
            # NOTE: I don't actually think this works for this function
            check_parts.append(self.value)

        # pre check for anything that needs to expand check parts
        for part in check_parts:
            if isinstance(part, HistoricalValue):
                check_parts.append(part.tracked_ref)

        for part in check_parts:
            if part in already_checked:
                continue
            already_checked.append(part)
            if isinstance(part, search_type):
                refs.append(part)

            refs.extend(part.find_refs_of_type(search_type, already_checked))
        return refs

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        """Get a pytensor graph representing this piece of an equation."""
        raise NotImplementedError(
            f"Failed at {str(self)}, no pt() implementation found for this equation part."
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        """Construct a string containing relevant pytensor code for this piece
        of the equation. This is useful for "compiling" into pymc code."""
        raise NotImplementedError(
            f"Failed at {str(self)}, no pt_str() implementation found for this equation part."
        )

    def is_static(self) -> bool:
        """Convenience shortcut for ``reno.utils.is_static()`` - True if this equation
        doesn't rely on any dynamic values (thus constant), False if it does."""
        return reno.utils.is_static(self)


# ==================================================
# PRIMITIVES
# ==================================================
# The base level things that can be used in an equation. Most of these
# are subclassed elsewhere (minus Scalar) and those subclasses are probably
# what someone using this API should be using.


class Scalar(EquationPart):
    """A static, single value equation part, representing some simple value that
    doesn't need to be computed.

    Args:
        value (int | float | np.ndarray): The scalar value to use.
    """

    def __init__(self, value: int | float | list | np.ndarray):
        super().__init__()
        if isinstance(value, list):
            value = np.array(value)
        self.value = value

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """No compute necessary, just get the previously specified value. (And
        likely use it in the rest of the equation.)"""
        return self.value

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        tensor = pt.as_tensor(self.value)
        # consolidate non-integer types to float64, otherwise pymc gets real mad
        if tensor.dtype == "float32":
            tensor = tensor.astype("float64")
        return tensor

    def pt_str(self, **refs: dict[str, str]) -> str:
        if pt.as_tensor(self.value).dtype == "float32":
            return f'pt.as_tensor({self.value}).astype("float64")'
        return f"pt.as_tensor({self.value})"

    def __repr__(self):
        return f"Scalar({self.value})"


class Distribution(EquationPart):
    """Represents a probability distribution or set that is drawn from for each
    sample (n) in a simulation.

    Probably shouldn't directly be using this, although you can in theory
    likely pass in ``partial()``s with numpy functions (just minus the size
    variable.)

    Args:
        func (Callable): The function that will be used to draw from.
    """

    def __init__(self):
        super().__init__()

    def populate(self, n: int):
        """Generate n samples based on this probability distribution, assigns
        as a vector to ``self.value``.

        Args:
            n (int): Number of samples to draw.
        """
        # should be implemented in child classes

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """Get the vector of samples pulled from the probability distribution.

        Note that this expects ``populate()`` to already have been called."""
        if row_indices is None:
            return self.value
        return self.value[row_indices]


class Operation(EquationPart):
    """Parent class to represent a mathematical operation in an equation.

    Pretty useless by itself, look in reno/ops.py, we use a parent class
    to make it easier to distinguish/find parts of an equation that are
    operations.

    Args:
        operands (list[EquationParts]): The sub equations the operation is
            applying across.
    """

    OP_REPR = None
    """If defined, this is what gets printed as the 'label' for the operation in the repr,
    and is what's used during parsing. Otherwise, just uses class name."""

    def __init__(self, *operands):
        super().__init__(list(operands))

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """'wrap' all operation eval functions so we get better error handling."""
        try:
            return self.op_eval(
                t=t, save=save, row_indices=row_indices, force=force, **kwargs
            )
        except Exception as e:
            e.add_note(
                f'Was evaluating operation "{self.__class__}" ({self.op_repr()}) with sub equation parts:'
            )
            for sub_eq in self.sub_equation_parts:
                e.add_note(f"\t{sub_eq}")
            raise

    def op_eval(self, **kwargs) -> int | float | np.ndarray:
        """Any new operations should implement this method. (Put all evaluation
        logic in here, as opposed to overriding eval.)"""
        raise NotImplementedError()

    @classmethod
    def op_repr(cls) -> str:
        """Get a string representation for the op name/label, used for
        printing and parsing."""
        if cls.OP_REPR is not None:
            return cls.OP_REPR
        # use the class name if no OP_REPR was set on the class.
        return cls.__name__

    @staticmethod
    def op_types() -> list[type]:
        """Get a list of classes that inherit from this one, in other words all of
        the possible defined operations."""
        subclasses = []  # collected list of types
        to_process = [Operation]  # the "frontier"
        while len(to_process) > 0:
            # get the next class to search
            parent_class = to_process.pop()

            # any new subclasses and make sure they're in the frontier
            for child_class in parent_class.__subclasses__():
                if child_class not in subclasses:
                    subclasses.append(child_class)
                    to_process.append(child_class)
        return subclasses

    def __repr__(self):
        sub_parts_string = " ".join(
            [subpart.__repr__() for subpart in self.sub_equation_parts]
        )
        return f"({self.op_repr()} {sub_parts_string})"


class Reference(EquationPart):
    """A symbolic reference to some other equation component, e.g. a
    stock, flow, or variable.

    This is a largely semantic parent class, probably shouldn't directly be
    using this.

    Semantically a reference is anything that will need to "resolve" somehow,
    either by looking up a historical value or running a sub equation to get
    the result.

    Args:
        label (str): The visual label for the reference. In the context of a model,
            is set to the name assigned on the model if not explicitly provided.
    """

    def __init__(self, label: str = None, doc: str = None):
        self.name = None
        self.label = label
        """Label is what's used in any visual representation (e.g. allows spaces where
        name does not.)"""
        self.doc = doc
        """A docstring to explain/describe the reference."""
        super().__init__()

    def latex(self, **kwargs) -> str:
        """String representation suitable for a latex display."""
        latex_str = latex_name(self.label)
        if hasattr(self, "model") and self.model.parent is not None:
            latex_str += f"_{{{latex_name(self.model.label)}}}"
        if "t" in kwargs:
            t = kwargs["t"]
            sample = kwargs["sample"]
            out_value = self.eval(t)
            if isinstance(out_value, (list, np.ndarray)):
                out_value = out_value[sample]

            latex_str += (
                "{\\color{grey}\\{}{\\color{red}"
                + str(out_value)
                + "}{\\color{grey}\\}}"
            )
        if "hl" in kwargs:
            if kwargs["hl"] == self.name or (
                hasattr(self, "qual_name") and kwargs["hl"] == self.qual_name()
            ):
                latex_str = "{\\color{cyan}" + latex_str + "}"

        return latex_str

    def __repr__(self):
        return f'"{self.label}"'


class Piecewise(EquationPart):
    """A conditional allowing evaluation of two or more condition equations
    to determine which output equation to use.

    Mathematically allows something like:
           ⎧ 1     if t < 4
    f(t) = ⎨ 5     if 4 <= t < 7
           ⎩ 9 + t if 7 <= t

    Args:
        equations (list[EquationPart]): The possible equation branches that
            selectively evaluate based on which condition is true. Must have
            same number of equations as conditions.
        conditions (list[Callable]): The boolean conditions to evaluate to
            determine which equation to output. These can all either be an
            equation based on EquationPart (see the boolean ops in ops.py),
            or a function that accepts a single parameter (t) as an input.
            Must have same number of conditions as equations.
    """

    # TODO: is this an operation? It kind of is isn't it? Not sure if it's
    # worth inheriting from operation or not, biggest difference I'd think
    # would be the op_eval vs eval (better debugging?) and less jank to include
    # piecewise handling in parser?
    # TODO: another option would be to support arbitrary *args and each
    # condition would follow each equation. This would likely get rid of
    # separate need for parse
    def __init__(self, equations: list[EquationPart], conditions: list[Callable]):
        # ensure scalar on each equation to allow specifying raw integers as
        # outputs without having to specify Scalar
        equations = [ensure_scalar(equation) for equation in equations]
        all_parts = [*equations]
        for condition in conditions:
            # if hasattr(condition, "eval"):
            if isinstance(condition, EquationPart):
                all_parts.append(condition)
        super().__init__(all_parts)
        self.equations = equations
        self.conditions = conditions

        assert len(self.equations) == len(self.conditions)

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """Evaluate condition equations/functions until a ``True`` is returned,
        and then evaluate and return the corresponding equation."""

        condition_bools = [
            condition.eval(t, save, row_indices, force, **kwargs)
            for condition in self.conditions
        ]

        shapes = [
            condition_bool.shape[0] if hasattr(condition_bool, "shape") else 1
            for condition_bool in condition_bools
        ]
        max_shape = max(shapes)

        # handle basic case where we're dealing with simple booleans
        if max_shape == 1:
            for index, condition_bool in enumerate(condition_bools):
                if condition_bool:
                    return self.equations[index].eval(
                        t, save, row_indices, force, **kwargs
                    )
        else:
            if row_indices is None:
                indices_array = np.arange(0.0, max_shape, 1.0)
            else:
                indices_array = np.asarray(row_indices).astype("float")

            funcs = []
            for equation in self.equations:
                # NOTE: we need the eqn=equation so that equation isn't just
                # always pointing to the last one in the list
                # https://stackoverflow.com/questions/7546285/creating-lambda-inside-a-loop
                funcs.append(
                    lambda x, eqn=equation: eqn.eval(
                        t, save, x.astype("int"), force, **kwargs
                    )
                )

            return np.piecewise(indices_array, condition_bools, funcs)

        raise Exception(  # pylint: disable=W0719
            f"Piecewise function {self.value} had no valid conditions."
        )

    def latex(self, **kwargs) -> str:
        """Get a string representation for the cool left bracket with the equations
        and their corresponding conditions. Conditions that aren't EquationParts can't
        really be converted meaningfully, so are represented simply with
        ``\\lambda(t)``"""
        # allow a more generic representation for limited latex environments,
        # like trying to output a text png from matplotlib
        # if "simplified_output" in kwargs and kwargs["simplified_output"]:
        #     pass
        # else
        string = "\\begin{cases}\n"
        for index, eq in enumerate(self.equations):
            condition_repr = (
                "\\lambda(t)"
                if not hasattr(self.conditions[index], "latex")
                else self.conditions[index].latex(**kwargs)
            )
            string += eq.latex(**kwargs) + " & \\text{if }" + condition_repr + " \\\\\n"
        string += "\\end{cases}"
        return string

    def pt_condition(self, i, refs) -> pt.TensorVariable:
        """Get the pytensor equation starting at the ith condition (using ifelse).
        Recursively includes all >i conditions."""
        condition = self.conditions[i].pt(**refs)
        eqn1 = self.equations[i].pt(**refs)
        eqn2 = self.equations[i + 1].pt(**refs)
        if i < len(self.conditions) - 2:
            eqn2 = self.pt_condition(i + 1, refs)
        # pymc gets the big angry if you try to mix float64 and float32
        # pymc getting the big angry gets me the big angry
        if eqn1.dtype == "float32":
            eqn1 = eqn1.astype("float64")
        if eqn2.dtype == "float32":
            eqn2 = eqn2.astype("float64")
        return ifelse(condition, eqn1, eqn2)

    def pt_condition_str(self, i, refs) -> str:
        """Get the string for the pytensor code for the ith condition (using ifelse).
        Recursively includes all >i conditions."""
        condition_str = self.conditions[i].pt_str(**refs)
        # we get the actual equations just so we can check if casts are necessary
        eqn1 = self.equations[i].pt()
        eqn1_str = self.equations[i].pt_str(**refs)
        eqn2 = self.equations[i + 1].pt()
        eqn2_str = self.equations[i + 1].pt_str(**refs)
        if i < len(self.conditions) - 2:
            eqn2 = self.pt_condition(i + 1, {})
            eqn2_str = self.pt_condition_str(i + 1, refs)
        if eqn1.dtype == "float32":
            eqn1_str += '.astype("float64")'
        if eqn2.dtype == "float32":
            eqn2_str += '.astype("float64")'
        return f"ifelse({condition_str}, {eqn1_str}, {eqn2_str})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.pt_condition(0, refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.pt_condition_str(0, refs)

    @staticmethod
    def parse(arg_strs: list[str], refs: dict[str, Reference]):
        """Parsing a piecewise has to be handled differently from normal operations because
        of the collection of associated arguments."""
        equations = []
        conditions = []
        for arg_str in arg_strs:
            arg_str_parts = reno.parser.parse_op_str(arg_str, no_op=True)
            if len(arg_str_parts) != 2:
                raise SyntaxError(
                    f"Every piecewise argument should contain a condition and an equation, found {len(arg_str_parts)} parts in '{arg_str}'"
                )
            condition = reno.parser.parse(arg_str_parts[0], refs)
            equation = reno.parser.parse(arg_str_parts[1], refs)
            conditions.append(condition)
            equations.append(equation)
        return Piecewise(equations, conditions)

    def __repr__(self):
        piece_str = "(piecewise"
        for i, condition in enumerate(self.conditions):
            piece_str += f" ({condition.__repr__()} {self.equations[i].__repr__()})"
        piece_str += ")"
        return piece_str


class Function(Reference):
    """Represents a call to a python function _outside_ of a model/system of
    equations.

    This is to allow for computations that might be a little complex, too fancy,
    or too annoying to specify entirely within the ``EquationPart`` system. By
    wrapping the call with this class, we defer actual computation/computation
    of input parameters until eval time, it's a thunk!

    Note:
        Any passed arguments that _are_ equations will be evaluated and passed
        in to the actual function call when ``.eval()`` is run. In a simulation
        run with stocks/flows etc., these arguments are likely to resolve to numpy
        arrays, which is the vector of all samples for that argument at the current
        eval'ed timestep. (So the function should expect to do vector operations,
        numpy will probably automagically handle this.)

    Args:
        f (Callable): The python function to run when ``.eval()`` is called.
        args (list[any]): Arguments to pass to ``f`` when ``.eval()`` is called.
            Any ``EquationPart`` arguments will have their ``.eval()`` called prior
            to ``f``'s execution.
        inject_mathlib (bool): Whether to pass a `mathlib` arg to allow running
            in both normal reno math mode as well as pymc. Requires underlying
            function to take `mathlib`.
        pt_mathlib: The library to use for pytensor operations.
        np_mathlib: The library to use for numpy operations.
        kwargs (list[any]): Keyword args to pass to ``f`` when ``.eval()`` is
            called. Any ``EquationPart`` arguments will have their ``.eval()``
            called prior to ``f``'s execution.

    NOTE: theoretically works in pymc conversion (as long as mathlib is used) but
    relatively untested as of writing.

    NOTE: pt_str isn't implemented in this class due to difficulty in referring
    to an unknown external function via a string.
    """

    # TODO: Possibly make this a TrackedReference and move to "complex" section
    # of this module, it would be useful to see a history of function outputs,
    # and the caching for expensive functions might be nice if the function gets
    # referenced (eval'd) multiple times.

    def __init__(
        self,
        f: Callable,
        *args,
        inject_mathlib: bool = True,
        pt_mathlib=None,
        np_mathlib=None,
        **kwargs,
    ):
        if pt_mathlib is None:
            pt_mathlib = pt
        if np_mathlib is None:
            np_mathlib = np

        self.f = f
        self.args = args
        self.kwargs = kwargs

        self.inject_mathlib = inject_mathlib
        self.pt_mathlib = pt_mathlib
        self.np_mathlib = np_mathlib

        # find any equation parts in the arguments so we can set the
        # sub_equation_parts of the parent class.
        eqn_parts = []
        for arg in args:
            if isinstance(arg, EquationPart):
                eqn_parts.append(arg)
        for kwarg in kwargs.values():
            if isinstance(kwarg, EquationPart):
                eqn_parts.append(kwarg)

        # TODO: ref should probably allow taking eqn parts?
        super().__init__(f.__name__)
        self.sub_equation_parts = eqn_parts

    def __repr__(self):
        return f"{self.f.__name__}()"

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """Eval any function arguments as needed, and then call the function itself.

        Args:
            t (int): Timestep along simulation at which to evaluate.
            save (bool): Ignored in this subclass, but passed along in sub-
                ``.eval()`` calls.
        """
        pass_args = []
        pass_kwargs = {}

        # handle mathlib if necessary
        if self.inject_mathlib:
            pass_kwargs["mathlib"] = self.np_mathlib

        # resolve any arguments that are EquationParts
        for arg in self.args:
            if isinstance(arg, EquationPart):
                pass_args.append(arg.eval(t, save, row_indices, force, **kwargs))
            else:
                pass_args.append(arg)
        for key, arg in self.kwargs.items():
            if isinstance(arg, EquationPart):
                pass_kwargs[key] = arg.eval(t, save, row_indices, force, **kwargs)
            else:
                pass_kwargs[key] = arg

        # call the actual function
        return self.f(*pass_args, **pass_kwargs)

    def latex(self, **kwargs) -> str:
        """Get a latex-suitable string representation of the function, use texttt to
        make it look like code and distinguishable from flows/vars/stocks/etc."""
        parameters = [part.latex(**kwargs) for part in self.sub_equation_parts]
        latex_str = f"{latex_name(self.label, 'texttt')}({', '.join(parameters)})"
        if "t" in kwargs:
            t = kwargs["t"]
            sample = kwargs["sample"]
            out_value = self.eval(t)
            if isinstance(out_value, (list, np.ndarray)):
                out_value = out_value[sample]
            latex_str += (
                "{\\color{grey}\\{}{\\color{yellow}"
                + str(out_value)
                + "}{\\color{grey}\\}}"
            )
        return latex_str

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        pass_args = []
        pass_kwargs = {}

        # handle mathlib if necessary
        if self.inject_mathlib:
            pass_kwargs["mathlib"] = self.pt_mathlib

        # resolve any arguments that are EquationParts
        for arg in self.args:
            if isinstance(arg, EquationPart):
                pass_args.append(arg.pt(**refs))
            else:
                pass_args.append(arg)
        for key, arg in self.kwargs.items():
            if isinstance(arg, EquationPart):
                pass_kwargs[key] = arg.pt(**refs)
            else:
                pass_kwargs[key] = arg

        # call the actual function
        return self.f(*pass_args, **pass_kwargs)

    # TODO: this is actually difficult because we don't know how the function is
    # being imported/thus don't know how to call it by name correctly?
    # we could potentially store in a global somehow, but this gets messy fast.
    # Leaving this here for now in case I get a brilliant idea for how to make
    # it work.
    # def pt_str(self, **refs: dict[str, str]) -> str:
    #     pass_args = []
    #     pass_kwargs = {}
    #
    #     # handle mathlib if necessary
    #     if self.inject_mathlib:
    #         pass_kwargs["mathlib"] = self.pt_mathlib
    #
    #     # resolve any arguments that are EquationParts
    #     for arg in self.args:
    #         if isinstance(arg, EquationPart):
    #             pass_args.append(arg.pt(**refs))
    #         else:
    #             pass_args.append(arg)
    #     for key, arg in self.kwargs.items():
    #         if isinstance(arg, EquationPart):
    #             pass_kwargs[key] = arg.pt(**refs)
    #         else:
    #             pass_kwargs[key] = arg
    #
    #     # call the actual function
    #     return self.f(*pass_args, **pass_kwargs)
    #


class TimeRef(Reference):
    """A reference to the current simulation timestep that can be used in
    equations.

    Example:
        >>> from reno.components import TimeRef
        >>> t = TimeRef()

        >>> t + 3
        (+ t 3)

        >>> (t + 3).eval(4)
        7
    """

    def __init__(self):
        super().__init__("t")
        self.name = "t"

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int:
        return t

    def latex(self, **kwargs) -> str:
        return "\\textit{t}"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        if self.name in refs:
            refs[self.name].name = self.name
            return refs[self.name]
        return pt.scalar(self.name)

    def pt_str(self, **refs: dict[str, str]) -> str:
        if self.name in refs:
            return refs[self.name]
        return f'pt.scalar("{self.name}")'

    def __repr__(self):
        return '"t"'


# ==================================================
# SYSTEM DYNAMICS FANCY COMPONENTS
# ==================================================
# The actual things used to define system dynamics models, generally these are
# components guaranteed to have one single sub equation that defines how they
# change. These components also have the option to track historical eval
# outputs.


class TrackedReference(Reference):
    """A reference that keeps a matrix of all its values over a simulation for
    caching and reporting purposes. This is the base class for the fancier
    components: Flow, Variable, and Stock.

    Probably should be using one of the subclasses of this.

    A reminder that when Flows and Vars are evaluated, they are based on the
    current timestep, e.g.:

    ```
    flow1(t) = flow2(t) + var1(t)
    ```

    As opposed to stocks, whose update equations are based on timestep t-1.
    This means that flows and vars should not have circular references (causing
    infinite recursion errors.)
    """

    def __init__(
        self,
        eq: EquationPart = None,
        label: str = None,
        doc: str = None,
        min: EquationPart = None,  # pylint: disable=W0622
        max: EquationPart = None,  # pylint: disable=W0622
        init: int | float | Distribution | Scalar | EquationPart = None,
        group: str = "",
    ):
        super().__init__(label, doc)
        self.eq = eq
        self.group = group

        self.min = ensure_scalar(min)
        self.max = ensure_scalar(max)

        # use this to prevent infinite recursion in the case of historical
        # values being used.
        self._computing = []

        self.model = None
        """Keep a reference to container model, makes it easier to compare refs across
        multiple models."""

        self.init: int | float | Distribution | Scalar | EquationPart = ensure_scalar(
            init
        )
        """Initial value/equation for initial value for stock/flow/var at t=0"""

        self._static: bool = False
        """If this is a reference that only needs to have a value calculated once and
        then never changes, don't track a full matrix for it.

        NOTE: this should only really be relevant for variables and possibly flows.

        Static effectively assumes this ref is a vector (or single value?)

        Note that this is currently being computed from the .populate() function.
        NOTE: _static is only meant for internal use for efficiency (to avoid constant
        round trips to utils.is_static()) during eval(), use is_static for all other
        cases.
        """
        self.static_value_computed: bool = False
        """If this is a static reference, we need to make sure the computation has
        run at least once, so flip this after the first time to signify that future
        evals can short-circuit."""

    def min_refs(self) -> list:
        """Get any references found in the min constraint equation. Currently mostly
        only used to aid in diagrams."""
        if self.min is None:
            return []
        if isinstance(self.min, TrackedReference):
            return [self.min]
        return self.min.seek_refs()

    def max_refs(self) -> list:
        """Get any references found in the max constraint equation. Currently mostly
        only used to aid in diagrams."""
        if self.max is None:
            return []
        if isinstance(self.max, TrackedReference):
            return [self.max]
        return self.max.seek_refs()

    def _implied_eq(self, eq=None):
        """The full equation that takes min/max equations into account."""

        # TODO: this doesn't handle if max < min or min > max
        implied_eq = self.eq if eq is None else eq
        if self.max is not None:
            implied_eq = reno.ops.minimum(implied_eq, self.max)
        if self.min is not None:
            implied_eq = reno.ops.maximum(implied_eq, self.min)

        return implied_eq

    def qual_name(self, dot: bool = False):
        """Get a string with both the model and the reference name
        if this model is a submodel of something else.

        This is primarily used for helping distinguish things in a
        multimodel setup."""
        if self.model is not None and self.model.parent is not None:
            delim = "_" if not dot else "."
            return f"{self.model.name}{delim}{self.name}"
        return self.name

    def populate(self, n: int, steps: int):
        """Initialize the matrix of values with size ``n x steps``. All
        values initially nan to indicate they need to be computed still

        Args:
            n (int): The number of samples to simulate.
            steps (int): How many steps will be run in the simulation.
        """
        # TODO: not sure if auto checking for staticness _here_ is the correct
        # place
        self.static_value_computed = False
        self._determine_if_static()
        if self._static:
            self.value = np.empty((n,))
        else:
            self.value = np.empty((n, steps))
        self.value.fill(np.nan)
        self.initial_vals()

    def _determine_if_static(self):
        """Used to help improve efficiency of static checks since done every
        single eval call."""
        self._static = (
            not isinstance(self, Stock)
            and is_static(self.eq)
            and is_static(self.min)
            and is_static(self.max)
        )

    def resolve_init_array(self, obj_or_eq):
        """Convert a number or scalar/distribution into correct
        starting array. Can be used within initial_vals subclass
        definition."""
        if isinstance(obj_or_eq, (int, float)):
            return obj_or_eq
        if isinstance(obj_or_eq, Distribution):
            obj_or_eq.populate(self.value.shape[0])
            return obj_or_eq.eval(0)
        if isinstance(obj_or_eq, EquationPart):
            # this covers Scalar
            return obj_or_eq.eval(0)
        return obj_or_eq

    def initial_vals(self):
        """Subclasses that have a special way of defining the initial set of
        values (e.g. stocks with their _0 values) can override this function,
        automatically called at the end of ``populate()``.
        """
        # define in subclass if relevant

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """Compute the equation for the given timestep. If we've already computed
        for this step (indicated by a value in the matrix that is not NaN), return
        that instead (unless told otherwise with ``force=True``).

        Note that evaluation output is only stored in the tracking matrix if
        save is ``True``, this class and subclasses are why save is being
        passed through all other ``.eval()`` methods.
        """
        try:
            # check to see if we can short-circuit computation as long as
            # a re-compute wasn't explicitly requested
            if not force:
                # short-circuit if this is a static vector value. (No computation necessary)
                if self._static and self.static_value_computed:
                    if row_indices is not None and len(self.value) > 1:
                        return self.value[row_indices]
                    return self.value

                # check to see if we've already computed this and short circuit if
                # so (we can do this now because we're basing uncomputed status
                # based on np.nan value rather than an internal step counter)
                # if not self.static: #and self.static_value_computed:
                if not self._static and self.value is not None:
                    if row_indices is not None:
                        if not np.isnan(self.value[row_indices, t]).any():
                            return self.value[row_indices, t]
                    if not np.isnan(self.value[:, t]).any():
                        return self.value[:, t]

            # short-circuit not appropriate, run actual compute
            self._computing.append(t)
            if isinstance(self, Stock):
                # Reminder that stocks are always the end value at t-1, while
                # flows and variables are current timestep computations. See
                # docstring of this class.
                val = self._implied_eq().eval(t - 1, save, row_indices, force, **kwargs)
                # TODO: if self is stock, don't pass save on down?
                # (should only be calling eval from model which is already handling
                # saving everything?)
            else:
                val = self._implied_eq().eval(t, save, row_indices, force, **kwargs)
            if save and not force:
                # if this is a static ref, now that we've computed the value we
                # can mark it for future short-circuiting.
                if self._static:
                    self.static_value_computed = True
                    if row_indices is not None:
                        self.value[row_indices] = val
                    else:
                        if not isinstance(val, np.ndarray):
                            self.value = np.array([val])
                        else:
                            self.value = val
                else:
                    if row_indices is None:
                        self.value[:, t] = val
                    else:
                        self.value[row_indices, t] = val
            self._computing.remove(t)
            return val
        except Exception as e:
            e.add_note(f'Was evaluating "{self.name}": {self._implied_eq()}')
            if self._implied_eq() is None:
                e.add_note(
                    f'\tIt looks like "{self.name}" was None? Make sure all refs have a value or equation'
                )
            raise

    def history(self, index_eq: EquationPart):
        """Get a reference to a previous value of this reference.

        Args:
            index_eq (EquationPart): An equation defining the timestep in this ref's
                historical values to use. Note that currently, pymc conversions only
                support static time-based equations, e.g. ``t - 3``.
        """
        return HistoricalValue(self, index_eq)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        if self.qual_name() in refs:
            if refs[self.qual_name()].name is None:
                refs[self.qual_name()].name = self.qual_name()
            return refs[self.qual_name()]
        return pt.scalar(self.qual_name())

    def pt_str(self, **refs: dict[str, str]) -> str:
        if self.qual_name() in refs:
            return refs[self.qual_name()]
        return f'pt.scalar("{self.qual_name()}")'

    def to_dict(self) -> dict:
        """Serialize class into a dictionary for saving to file."""
        return {
            "label": self.label,
            "init": str(self.init),
            "doc": self.doc,
            "min": str(self.min),
            "max": str(self.max),
            "group": self.group,
            "eq": str(self.eq),
        }

    def from_dict(self, data: dict, refs: dict[str, "TrackedReference"]):
        """Deserialize reference and parse data from dictionary previously saved from
        ``to_dict()``."""
        self.label = data["label"]
        self.init = reno.parser.parse(data["init"], refs)
        self.doc = data["doc"]
        self.min = reno.parser.parse(data["min"], refs)
        self.max = reno.parser.parse(data["max"], refs)
        self.group = data["group"]
        if "eq" in data:
            self.eq = reno.parser.parse(data["eq"], refs)

    def __repr__(self):
        return f'"{self.qual_name()}"'


class HistoricalValue(Reference):
    """A wrapper class for a reference, specifically for getting a previous value indexed
    by some other equation."""

    def __init__(self, tracked_ref: TrackedReference, index_eq: EquationPart):
        super().__init__(label=tracked_ref.label)
        self.index_eq = index_eq
        self.tracked_ref = tracked_ref

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force=False,
        **kwargs,
    ):
        if t not in self.tracked_ref._computing:
            # the conditional is to prevent infinite recursion.
            # This piece is _really_ important, otherwise delays can mean
            # that certain timesteps for a variable/flow are never computed,
            # leading to inaccuracies and confusing table outputs. (esp if
            # an equation only ever refers to some variable's _history_, and
            # the variable itself (at timestep t) is never used)
            # So, make sure that the actual output of the actual flow/variable
            # is always "up to date"
            self.tracked_ref.eval(t, save, row_indices, force, **kwargs)

        # we ensure the reference has actually evaluated because otherwise
        # you might have a bunch of random zeros near the end of a simulation

        index = self.index_eq.eval(t, save, row_indices, force, **kwargs)
        if index < 0:
            return 0  # TODO: TODO: TODO: or initial condition?

        if self.tracked_ref._static:
            if row_indices is None:
                return self.tracked_ref.value
            return self.tracked_ref.value[row_indices]
        if row_indices is None:
            return self.tracked_ref.value[:, index]
        return self.tracked_ref.value[row_indices, index]

    def latex(self, **kwargs) -> str:
        """Get the string representation for referring to this reference, italicized
        and as a function of ``t`` to highlight it's a different timestep"""
        latex_str = (
            f"{latex_name(self.label, 'textit')}({self.index_eq.latex(**kwargs)})"
        )
        if "t" in kwargs:
            t = kwargs["t"]
            sample = kwargs["sample"]
            out_value = self.eval(t)
            if isinstance(out_value, (list, np.ndarray)):
                out_value = out_value[sample]
            latex_str += (
                "{\\color{grey}\\{}{\\color{red}"
                + f"{out_value:.2f}"
                + "}{\\color{grey}\\}}"
            )
        if "hl" in kwargs and kwargs["hl"] == self.tracked_ref.name:
            latex_str = "{\\color{cyan}" + latex_str + "}"
        return latex_str

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        simple_index = self.index_eq.eval(t=0, force=True) * -1
        name = f"{self.tracked_ref.qual_name()}_h{simple_index}"
        if name in refs:
            if refs[name].name is None:
                refs[name].name = name
            return refs[name]
        return pt.as_tensor(0.0)  # ???

    def pt_str(self, **refs: dict[str, str]) -> str:
        simple_index = self.index_eq.eval(t=0, force=True) * -1
        name = f"{self.tracked_ref.qual_name()}_h{simple_index}"
        if name in refs:
            return refs[name]
        return "pt.as_tensor(0.0)"  # ???

    @staticmethod
    def parse(arg_strs: list[str], refs: dict[str, Reference]):
        if len(arg_strs) != 2:
            raise SyntaxError(
                "history op expects two arguments, the tracked reference, and the indexing equation."
            )
        tracked_ref = reno.parser.parse(arg_strs[0], refs)
        index_eq = reno.parser.parse(arg_strs[1], refs)
        return HistoricalValue(tracked_ref, index_eq)

    def __repr__(self):
        return f"(history {self.tracked_ref.__repr__()} {self.index_eq.__repr__()})"


class Flow(TrackedReference):
    """A flow is a rate, a movement of some material/thing/quantity per some
    unit of time.

    Flow equations probably reference stocks, moving stuff from one stock
    to another etc. Flows are what stocks are defined in terms of (in_flows
    and out_flows.)

    https://insightmaker.com/docs/flows
    https://en.wikipedia.org/wiki/Stock_and_flow
    """

    def __init__(
        self,
        eq: EquationPart = None,
        label: str = None,
        doc: str = None,
        min: EquationPart = None,
        max: EquationPart = None,
        group: str = "",
    ):
        super().__init__(eq, label, doc, min=min, max=max, group=group)

    def initial_vals(self):
        """Variables can be set to distributions, so the inital vals
        in that case will be a population sampled from the eq distribution."""
        try:
            if self.init is not None:
                self.value[:, 0] = self.resolve_init_array(self._implied_eq(self.init))
            elif isinstance(self.eq, (Distribution, Scalar, int, float)):
                if self._static:
                    self.value = self.resolve_init_array(self._implied_eq(self.eq))
                else:
                    self.value[:, 0] = self.resolve_init_array(
                        self._implied_eq(self.eq)
                    )
            else:
                if self._static:
                    # self.value = np.array(0)
                    self.value = self.resolve_init_array(self._implied_eq())
                else:
                    self.value[:, 0] = self.resolve_init_array(self._implied_eq())
        except Exception as e:
            e.add_note(f'Was attempting to compute initial values for "{self.name}"')
            raise

    def equation(self, **kwargs) -> str:
        """Return the full latex string representation for the entire equation,
        not just the right hand side."""

        lhs = f"{latex_name(self.label, 'textit')}"
        if self.model.parent is not None:
            lhs += f"_{{{latex_name(self.model.label)}}}"
        lhs += "(t)"

        if self.eq is None:
            return f"{lhs} = \\texttt{{None}}"
        return f"{lhs} = {self.eq.latex(**kwargs)}" + range_eq_latex(
            self.min, self.max, **kwargs
        )

    def latex(self, **kwargs) -> str:
        """Get the string representation for referring to this flow, italized
        and as a function of ``t`` to better represent that this should be a
        rate?"""
        latex_str = f"{latex_name(self.label, 'textit')}"
        if self.model.parent is not None:
            latex_str += f"_{{{latex_name(self.model.label)}}}"
        latex_str += "(t)"
        if "t" in kwargs:
            t = kwargs["t"]
            sample = kwargs["sample"]
            out_value = self.eval(t)
            if isinstance(out_value, (list, np.ndarray)):
                out_value = out_value[sample]
            latex_str += (
                "{\\color{grey}\\{}{\\color{red}"
                + f"{out_value:.2f}"
                + "}{\\color{grey}\\}}"
            )
        if "hl" in kwargs and kwargs["hl"] == self.qual_name():
            latex_str = "{\\color{cyan}" + latex_str + "}"
        return latex_str

    def display(self):
        """Display an ipython markdown with the equation latex."""
        # TODO: _repr_latex
        # https://github.com/jupyter/ngcm-tutorial/blob/master/Part-1/IPython%20Kernel/Custom%20Display%20Logic.ipynb
        display(Markdown(f"${self.equation()}$"))

    def debug_equation(self, t, sample=0, **kwargs):
        return (
            self.latex(t=t, sample=sample, **kwargs)
            + " = "
            + self.eq.latex(t=t, sample=sample, **kwargs)
            + range_eq_latex(self.min, self.max, t=t, sample=sample, **kwargs)
        )

    def __setattr__(self, name, value):
        if name == "eq" and value is not None:
            value = ensure_scalar(value)
            # assign ops are necessary when setting to a single 'thing' rather
            # than an equation. Without this, seek_refs becomes almost
            # impossible to implement correctly.
            if isinstance(value, (Variable, Flow, Stock, HistoricalValue)):
                value = reno.ops.assign(value)
            # Make sure sub_equation_parts is kept up to date with the equation.
            self.sub_equation_parts = [value]
        elif name in ("min", "max", "init"):
            value = ensure_scalar(value)
        object.__setattr__(self, name, value)


class Variable(TrackedReference):
    """A variable is a static value(s) or function that can be used as part
    of other equations, e.g. flow definitons.

    https://insightmaker.com/docs/variables
    """

    def __init__(
        self,
        eq: EquationPart = None,
        label: str = None,
        doc: str = None,
        min: EquationPart = None,
        max: EquationPart = None,
        user: bool = False,
        group: str = "",
    ):
        self.user = user
        """If True, use visual interface to allow changing it via widgets."""
        # TODO: we're not currently using this in explorer, add that.

        if isinstance(eq, (float, int)):
            eq = Scalar(eq)

        super().__init__(eq, label, doc, min=min, max=max, group=group)

    def initial_vals(self):
        """Variables can be set to distributions, so the inital vals
        in that case will be a population samled from the eq distribution."""
        try:
            if self.init is not None:
                self.value[:, 0] = self.resolve_init_array(self._implied_eq(self.init))
            elif isinstance(self.eq, (Distribution, Scalar, int, float)):
                if self._static:
                    self.value = self.resolve_init_array(self._implied_eq(self.eq))
                else:
                    self.value[:, 0] = self.resolve_init_array(
                        self._implied_eq(self.eq)
                    )
            else:
                if self._static:
                    # self.value = np.array(0)
                    self.value = self.resolve_init_array(self._implied_eq())
                else:
                    self.value[:, 0] = self.resolve_init_array(self._implied_eq())
        except Exception as e:
            e.add_note(f'Was attempting to compute initial values for "{self.name}"')

    def equation(self, **kwargs) -> str:
        """Get the representation of the full equation for a variable as a latex
        string."""
        lhs = f"{latex_name(self.label)}"
        if self.model.parent is not None:
            lhs += f"_{{{latex_name(self.model.label)}}}"

        if self.eq is None:
            return f"{lhs} = \\texttt{{None}}"
        return f"{lhs} = {self.eq.latex(**kwargs)}" + range_eq_latex(
            self.min, self.max, **kwargs
        )

    def debug_equation(self, t, sample=0, **kwargs):
        return (
            self.latex(t=t, sample=sample, **kwargs)
            + " = "
            + self.eq.latex(t=t, sample=sample, **kwargs)
            + range_eq_latex(self.min, self.max, t=t, sample=sample, **kwargs)
        )

    def __setattr__(self, name, value):
        if name == "eq" and value is not None:
            value = ensure_scalar(value)
            # assign ops are necessary when setting to a single 'thing' rather
            # than an equation. Without this, seek_refs becomes almost
            # impossible to implement correctly.
            if isinstance(value, (Variable, Flow, Stock, HistoricalValue)):
                value = reno.ops.assign(value)
            # Make sure sub_equation_parts is kept up to date with the equation.
            self.sub_equation_parts = [value]
        elif name in ("min", "max", "init"):
            value = ensure_scalar(value)
        object.__setattr__(self, name, value)


class Stock(TrackedReference):
    """A stock represents some bucket or quantity of material/thing that
    can accumulate (in_flows) or deplete (out_flows) over time.

    https://en.wikipedia.org/wiki/Stock_and_flow
    https://insightmaker.com/docs/stocks

    Note that stock update equations are based on the previous timestep's values
    for all references (as opposed to flows/vars), e.g.:

    ```
    stock(t) = stock(t-1) + in_flows(t-1) - out_flows(t-1)
    ```
    """

    def __init__(
        self,
        label: str = None,
        init: int | float | Distribution | Scalar | EquationPart = None,
        doc: str = None,
        min: EquationPart = None,
        max: EquationPart = None,
        group: str = "",
    ):
        super().__init__(label, doc=doc, min=min, max=max, init=init, group=group)

        self.in_flows: list[Flow] = []
        self.out_flows: list[Flow] = []

    # ---- MATH OVERLOADING ----
    # this makes it so you can semantically define inflows/outflows with += and -=
    # e.g.
    # bath_water = Stock("bathtub")
    # bath_water += Flow("faucet")
    # bath_water -= Flow("drain")

    def __iadd__(self, obj):
        self.in_flows.append(obj)
        return self

    def __isub__(self, obj):
        self.out_flows.append(obj)
        return self

    # ---- /MATH OVERLOADING ----

    def initial_vals(self):
        try:
            if self.init is not None:
                self.value[:, 0] = self.resolve_init_array(self._implied_eq(self.init))
            else:
                self.value[:, 0] = 0
        except Exception as e:
            e.add_note(f'Was attempting to compute initial values for "{self.name}"')
            raise
        self.eq = self.compute_diff_eq()

    def equations(self, **kwargs) -> list[str]:
        """Get a list of string latex representations for all in and out flows."""

        lhs = f"{latex_name(self.label)}"
        if self.model.parent is not None:
            lhs += f"_{{{latex_name(self.model.label)}}}"

        eqns = []
        if len(self.in_flows) > 0:
            eqns.append(
                f"{lhs} \\mathrel{{{{+}}{{=}}}} {self.combine_eqs(self.in_flows).latex(**kwargs)}"
            )
        if len(self.out_flows) > 0:
            eqns.append(
                f"{lhs} \\mathrel{{{{-}}{{=}}}} {self.combine_eqs(self.out_flows).latex(**kwargs)}"
            )
        if self.min is not None or self.max is not None:
            eqns.append(f"{lhs}{range_eq_latex(self.min, self.max, **kwargs)}")
        return eqns

    def debug_equation(self, t, sample=0, **kwargs):
        return (
            self.latex(
                t=t + 1, sample=sample, **kwargs
            )  # TODO: minusone - prob need to change?
            + " = "
            + self.latex(t=t, sample=sample, **kwargs)
            + " + "
            + self.compute_diff_eq().latex(t=t, sample=sample, **kwargs)
        )

    def display(self):
        """Display a markdown with latex for inflow equations and outflow equations."""
        for eqn in self.equations():
            display(Markdown(f"${eqn}$"))

    def combine_eqs(self, eqs: list[EquationPart]) -> EquationPart:
        """Helper function to convert a list of equations into a single combined equation.
        (Essentially a big summation operator.)"""
        # TODO: could be moved out
        if len(eqs) == 0:
            return Scalar(0)
        if len(eqs) == 1:
            return eqs[0]
        eq = eqs[0]
        for equation in eqs[1:]:
            eq = eq + equation
        return eq

    def compute_diff_eq(self) -> EquationPart:
        """Create a combined equation that accounts for both inflows and outflows."""
        inflows = self.combine_eqs(self.in_flows)
        outflows = self.combine_eqs(self.out_flows)

        return inflows - outflows

    def _implied_eq(self, eq=None):
        """Overriding parent to automatically add the self + self.eq"""
        if eq is None:
            eq = self + self.compute_diff_eq()
        return super()._implied_eq(eq=eq)

    def plot(self, ax=None, plot_kwargs=None):
        """Generate a matplotlib plot for this stock's simulation."""
        # TODO: to reduce necessary dependencies of core component classes,
        # should maybe move plotting stuff into functions in separate module.
        # TODO: more specifically, diagrams add_stocks is still depending
        # on this, but should be able to switch to plot_trace_refs?
        if plot_kwargs is None:
            plot_kwargs = {}
        if ax is None:
            fig, ax = plt.subplots()
        if self.value is not None:
            for row in range(self.value.shape[0]):
                ax.plot(self.value[row])
        return ax

    def to_dict(self) -> dict:
        """Serialize class into a dictionary for saving to file. Stock has to modify
        the parent TrackedReference class serialization to account for equations being
        handled a little differently."""
        tracked_ref_dict = super().to_dict()
        del tracked_ref_dict["eq"]
        tracked_ref_dict["in_flows"] = [str(flow) for flow in self.in_flows]
        tracked_ref_dict["out_flows"] = [str(flow) for flow in self.out_flows]
        return tracked_ref_dict

    def from_dict(self, data: dict, refs: dict[str, TrackedReference]):
        """Deserialize reference and parse data from dictionary previously saved from
        ``to_dict()``."""
        super().from_dict(data, refs)
        for in_flow in data["in_flows"]:
            self.in_flows.append(reno.parser.parse(in_flow, refs))
        for out_flow in data["out_flows"]:
            self.out_flows.append(reno.parser.parse(out_flow, refs))

    def __setattr__(self, name, value):
        # TODO: this is mostly the same across all tracked refs, can prob move
        # to parent class
        if name == "min" or name == "max" or name == "init":
            value = ensure_scalar(value)
        object.__setattr__(self, name, value)


# ==================================================
# MEASUREMENTS
# ==================================================


class Metric(Reference):
    """Metrics run in a separate after-simulation analysis."""

    def __init__(self, eq: EquationPart = None, label: str = None):
        super().__init__(label)
        self.eq = eq
        self.model = None
        """Keep a reference to container model, makes it easier to compare refs across
        multiple models."""

    def equation(self, **kwargs) -> str:
        """Get the representation of the full equation for the metric as a latex
        string."""
        if self.eq is None:
            return f"{latex_name(self.label)} = \\texttt{{None}}"
        return f"{latex_name(self.label)} = {self.eq.latex(**kwargs)}"

    def debug_equation(self, t, sample=0, **kwargs):
        return (
            self.latex(t=t, sample=sample, **kwargs)
            + " = "
            + self.eq.latex(t=t, sample=sample, **kwargs)
        )

    def qual_name(self, dot: bool = False):
        """Get a string with both the model and the reference name
        if this model is a submodel of something else.

        This is primarily used for helping distinguish things in a
        multimodel setup."""
        if self.model is not None and self.model.parent is not None:
            delim = "_" if not dot else "."
            return f"{self.model.name}{delim}{self.name}"
        return self.name

    def to_dict(self) -> dict:
        """Serialize class into a dictionary for saving to file."""
        return {
            "label": self.label,
            "eq": str(self.eq),
        }

    def from_dict(self, data: dict, refs: dict[str, "TrackedReference"]):
        """Deserialize reference and parse data from dictionary previously saved from
        ``to_dict()``"""
        self.label = data["label"]
        self.eq = reno.parser.parse(data["eq"], refs)

    def __repr__(self):
        return f'"{self.qual_name()}"'


class PostMeasurement(Metric):
    """An equation to run after a simulation has complete, generally intended to
    be a series aggregate type equation (e.g. using series_min/series_max/sum on slices
    of values across time from references computed during the simulation etc."""

    def __setattr__(self, name, value):
        if name == "eq" and value is not None:
            if isinstance(value, (Variable, Flow, Stock, HistoricalValue)):
                value = reno.ops.assign(value)
            self.sub_equation_parts = [value]
        object.__setattr__(self, name, value)

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ):
        try:
            value = self.eq.eval(t, save, row_indices, force, **kwargs)
            if save:
                self.value = value
            return value
        except Exception as e:
            e.add_note(f'Was evaluating metric "{self.name}": {self.eq}')
            raise
        return value

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        if self.qual_name() in refs:
            if refs[self.qual_name()].name is None:
                refs[self.qual_name()].name = self.qual_name()
            return refs[self.qual_name()]
        return pt.scalar(self.qual_name())

    def pt_str(self, **refs: dict[str, str]) -> str:
        if self.qual_name() in refs:
            return refs[self.qual_name()]
        # return self.eq.pt_str(**refs)
        return f'pt.scalar("{self.qual_name()}")'


class Flag(Metric):
    """Boolean value tracked for each step and sample in the simulation."""

    def __init__(self, eq: EquationPart = None, label: str = None):
        super().__init__(eq, label)
        self.internal_step = 0
        self._computing = []
        # TODO: probably need to revise to similar to trackedrefs for how
        # computing is handled?

        # TODO: need to update first function to support pt as well
        self.first = Reference(f"{self.label}.first")
        self.first.eval = lambda t, save, row_indices, force: self.first.eq.eval(
            t, save, row_indices, force
        )
        self.first.eq = Function(self.first_event)

    def populate(self, n: int, steps: int):
        # TODO: TODO: this needs to act the same way as TrackedReference, using
        # NaN's instead of an internal step
        self.value = np.zeros((n, steps))
        self.internal_step = 0

    def eval(
        self,
        t: int,
        save: bool = False,
        row_indices: list[int] | np.ndarray = None,
        force: bool = False,
        **kwargs,
    ) -> int | float | np.ndarray:
        """If the timestep is less than our internal_step tracker, return the
        corresponding previous column in the tracking matrix, since that
        timestep has already been computed. Otherwise, compute, store, and
        update the internal step to t+1.

        Note that evaluation output is only stored in the tracking matrix if
        save is ``True``, this class and subclasses are why save is being
        passed through all other ``.eval()`` methods.
        """

        self._computing.append(t)

        # TODO: if eq is just a scalar or distribution, don't do this,
        # no need to store a bunch of rows/cols of a guaranteed static thing.

        # check if already computed
        # TODO: don't think caching works for now?
        # if self.internal_step > t:
        #     self._computing.remove(t)
        #     return self.value[:, t]
        val = self.eq.eval(t, save, row_indices, force, **kwargs).astype(int)
        if save:
            self.internal_step = t
            if row_indices is None:
                self.value[:, t] = val
            else:
                self.value[row_indices, t] = val
            self.internal_step += 1
        self._computing.remove(t)
        return val

    def indices(self):
        """Get the timesteps where the value changes from 0 to 1."""
        sample_indices, t_indices = np.where(np.diff(self.value) == 1)
        t_indices = t_indices + 1

        indices = []
        for i in range(self.value.shape[0]):
            i_indices = t_indices[np.where(sample_indices == i)[0]]
            if i_indices.shape[0] == 0:
                i_indices = np.asarray([np.nan])
            indices.append(i_indices)
        return indices

    def first_event(self):
        """Get the timestep for the first time the value is 1."""
        indices = self.indices()
        firsts = [sample_indices[0] for sample_indices in indices]

        return np.asarray(firsts)



"""Functions for translating a system dynamics model into a PyMC equivalent.

Because an SDM run here is a sequential simulation, the core enabling
functionality is the ``scan`` function in PyTensor:
https://pytensor.readthedocs.io/en/latest/library/scan.html
An example for how ``scan`` ties into PyMC:
https://www.pymc.io/projects/examples/en/latest/time_series/Time_Series_Generative_Graph.html

The output generated by these translations follows a template that looks
something like:

.. code-block:: python

    def step_function(*all_stocks_flows_variables):
        # this is generated by pt_sim_step() or pt_sim_step_str()

        s, ... = all_stocks_flows_variables

        # difference/update equations for each stock/flow/variable
        s_next = s + ...
        ...

        # return all updated values (pymc/pytensor has specific boilerplate to use)
        return [s_next, ...], ...

    with pymc.Model() as model:
        # define all pymc distributions/variables and initial stock/flow values
        # e.g. for stock "s":
        s_init = pm.Deterministic("s_init", ...)
        ...

        # run the scan function to generate full sequence data for simulation
        [s_seq, ...] = pytensor.scan(fn=step_function, [s_init, ...], ...)

        # create pymc containers for the full sequence data (this way they get
        # added into any output arviz datasets from pymc sampling operations)
        s = pm.Deterministic("s", pt.concatenate([s_init, s_seq], ...)

Reno's underlying math framework works fairly similarly to PyTensor, (symbolic
tree and numpy-ish) so translation is done by calling the ``pt()`` or ``pt_str()``
function that every EquationPart has, which returns the PyTensor equivalent of
that operation/thing.

Note that there are functions to directly create pymc models/operations and functions
to create strings of PyMC/PyTensor code (suffixed with _str). These exist for two reasons:

1. Debugging - figuring out errors or issues with PyMC models can be an adventure,
    and doubly so if you can't directly see or modify the model code. ``to_pymc_model_str()``
    should _in theory_ generate a string of code equivalent to what's used in
    ``to_pymc_model()``, but can be edited and run separately as needed to test
    changes.
2. Flexibility - if trying to do something beyond what Reno's PyMC naive translation
    can support, the string code output gives a starting point that can be directly
    incorporated or modified to do more advanced things.
"""

import warnings
from collections.abc import Callable

import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt

import reno


def pt_sim_step(model: "reno.model.Model") -> Callable:
    """Returns a target function for pytensor scan, this equates to a single step in the
    system dynamics model simulation. This probably shouldn't be used anywhere
    outside of the context of the pymc model creation function."""

    # general flow is to:
    # 1. run all stock update equations (which are based on previous values,
    #   being passed in through args)
    # 2. run dependency ordered non-static flow and variable equations
    #   (which are based on current values, hence the need for dependency
    #   ordering)

    def step(*args):
        original_refs = resolve_pt_scan_args(model, args)
        # keeping a non-modified version of refs to help with typechecking
        # at the end of this function
        refs = resolve_pt_scan_args(model, args)

        stock_nexts = []
        other_nexts = {}
        # other_nexts is dictionary because we need to reorder to default
        # flows + vars for returns, see note:
        # NOTE: assumed order of variables whenever they're being passed or returned
        # is order found in model.stocks + model.flows + model.vars

        for stock in model.all_stocks():
            full_stock_eq = stock._implied_eq(stock + stock.compute_diff_eq())
            stock_nexts.append(full_stock_eq.pt(**refs))

        # make sure further references to stocks (in the flows and vars) use
        # this timestep's stock computation result
        for index, stock in enumerate(model.all_stocks()):
            refs[stock.qual_name()] = stock_nexts[index]

        # dependency ordering for non-static flow/var eqs
        ref_compute_order = model.dependency_compute_order(inits_order=False)

        for obj in ref_compute_order:
            if isinstance(obj, reno.components.Stock):
                # no need to do stocks, already handled above
                continue
            if not obj.is_static():
                pt_eq = obj._implied_eq().pt(**refs)
                other_nexts[obj] = pt_eq
                refs[obj.qual_name()] = pt_eq

        # do any type checking to ensure consistency (if ints become floats without
        # an explicit conversion, PyTensor gets real unhappy)
        for index, stock in enumerate(model.all_stocks()):
            if stock_nexts[index].dtype != original_refs[stock.qual_name()].dtype:
                stock_nexts[index] = stock_nexts[index].astype(
                    original_refs[stock.qual_name()].dtype
                )
        for obj in ref_compute_order:
            if isinstance(obj, reno.components.Stock):
                # already handled in for loop above
                continue
            if not obj.is_static():
                if other_nexts[obj].dtype != original_refs[obj.qual_name()].dtype:
                    other_nexts[obj] = other_nexts[obj].astype(
                        original_refs[obj.qual_name()].dtype
                    )

        # order returns according to model.stocks + model.flows + model.vars
        returns = stock_nexts
        for obj in model.all_flows() + model.all_vars():
            if obj in other_nexts:
                returns.append(other_nexts[obj])

        return list(returns), pm.pytensorf.collect_default_updates(
            inputs=args, outputs=returns
        )

    return step


def pt_sim_step_imports() -> str:
    """Get the string of code for the imports needed for ``pt_sim_step_str()`` to
    run correctly."""
    import_strs = [
        "import pytensor.tensor as pt",
        "from pytensor.ifelse import ifelse",
        "import pymc as pm",
    ]
    return "\n".join(import_strs)


def pt_sim_step_str(model: "reno.model.Model") -> str:
    """Construct a string of python code for the step function for this model that could
    be used by a pytensor scan operation. Should be a functional (string) equivalent of
    the ``pt_sim_step()`` function.

    Expected imports for the resulting code to run:
        >>> import pytensor.tensor as pt
        >>> from pytensor.ifelse import ifelse
        >>> import pymc as pm

    (use ``pt_sim_step_imports()`` to get these programatically)

    Note that this code is in no way PEP8 formatted, for complex models some of the lines
    will be exceptionally long.
    """
    name = "step" if model.name is None else f"{model.name}_step"
    code = f"def {name}(*args):\n"
    code += f"\t{', '.join(expected_pt_scan_arg_names(model))} = args\n"

    refs_dict = expected_pt_scan_arg_names(model, as_dict=True)
    # (in pt_str calls, variables passed in from args should simply be substituted
    # with the variable name itself)

    code += "\n\t# Difference/recurrence equations\n"

    for stock in model.all_stocks():
        code += (
            f"\t{stock.qual_name()}_next = {stock._implied_eq().pt_str(**refs_dict)}\n"
        )

    # update all stock names to _next for use in flow/var updates
    # (we don't do this above on the off chance we somehow end up with a
    # stock being able to directly ref another stock? We'd want it to still
    # be based on old value at that point)
    for stock in model.all_stocks():
        refs_dict[stock.qual_name()] = f"{stock.qual_name()}_next"

    # dependency ordering for non-static flow/var eqs
    ref_compute_order = model.dependency_compute_order(inits_order=False)
    for obj in ref_compute_order:
        if isinstance(obj, reno.components.Stock):
            # no need to do stocks, already handled above
            continue
        if not obj.is_static():
            code += (
                f"\t{obj.qual_name()}_next = {obj._implied_eq().pt_str(**refs_dict)}\n"
            )
            refs_dict[obj.qual_name()] = f"{obj.qual_name()}_next"

    # type checking
    code += "\n\t# Type checks\n"
    for obj in ref_compute_order:
        if not obj.is_static():
            code += f"\t{obj.qual_name()}_next = {obj.qual_name()}_next.astype({obj.qual_name()}.dtype) if {obj.qual_name()}_next.dtype != {obj.qual_name()}.dtype else {obj.qual_name()}_next\n"

    # order returns according to model.stocks + model.flows + model.vars
    return_names = ", ".join(
        [f"{obj.qual_name()}_next" for obj in model.all_refs() if not obj.is_static()]
    )

    code += f"\n\treturn [{return_names}], pm.pytensorf.collect_default_updates(inputs=args, outputs=[{return_names}])\n"

    return code


def to_pymc_model(
    model: "reno.model.Model", observations: list["reno.ops.Observation"] = None
) -> pm.model.core.Model:
    """Generate a pymc model for bayesian analysis of this system dynamics model. The general
    idea is that this creates corresponding pymc variables (or distributions as relevant) for
    each stock/flow/var in the model, and sets up the full simulation sequence computations
    based on the generated step function from ``pt_sim_step()``.

    Sampling with priors should be equivalent to running the system dynamics model normally
    (this is essentially "forward simulation mode".) Add observations to the pymc model
    variables and sample from posterior to run bayesian analysis/determine how distributions
    of any other variables may be affected.
    """

    # general flow is to:
    # 1. set up all initial value computations in pymc deterministics (or
    #   appropriate distributions for reno.ops.distribution values) Note
    #   that _all_ initial computations need to be dependency ordered
    # 2. apply the pytensor scan function targeting ``pt_sim_step()``, any
    #   static references are passed as non-sequences, time is passed as
    #   a sequence, and all dynamic refs (stocks, flows, vars) are passed
    #   as outputs info, representing the initial values for each sequence
    #   and the set of expected return sequences
    # 3. create a deterministic with dim=timesteps for each output sequence
    #   from the scan operation.

    timesteps = model.steps
    coords = {"t": range(timesteps)}
    with pm.Model(coords=coords) as m:
        ref_compute_order = model.dependency_compute_order(inits_order=True)

        refs = {}
        inits_by_obj = {}
        # same logic as pt_sim_step for these being dictionaries, ordering is tricky with these,
        # we always assume arg/return orders are in default stocks + flows + vars order
        statics_by_obj = {}
        # note that we separate "statics" because the pytensor scan
        # equivalent is to pass them as "non-sequences" (they're passed in a
        # separate portion of the *args, and assumed to not update/not
        # requiring a return)
        hist_vars_by_obj = {}
        # any variables that require special treatment for historical
        # values have their starting historical array deterministics
        # stored in here

        refs["__PT_SEQ_LEN__"] = pt.as_tensor(model.steps)
        # there are a few operations that need to know the total length
        # of the simulation/sequence in order to create the correct
        # pytensor equivalent output, see ops.py

        historical_ref_taps = find_historical_tracked_refs(model)

        # have an initial t=0 variable if necessary
        if model.find_timeref_name() is not None:
            refs[model.find_timeref_name()] = pt.as_tensor(0)

        # create symbolic computations for all initial values for stocks/flows/vars
        for obj in ref_compute_order:
            if isinstance(obj, reno.components.Stock) and obj.init is None:
                pt_eq = pm.Deterministic(
                    f"{obj.qual_name()}_init", pt.as_tensor(np.array(0.0))
                )
                inits_by_obj[obj] = pt_eq
                refs[obj.qual_name()] = pt_eq
            elif obj.is_static():
                if isinstance(obj.eq, reno.components.Distribution):
                    pt_eq = obj.eq.pt(__PTNAME__=obj.qual_name(), **refs)
                else:
                    pt_eq = pm.Deterministic(
                        obj.qual_name(), obj._implied_eq(obj.eq).pt(**refs)
                    )
                statics_by_obj[obj] = pt_eq
                refs[obj.qual_name()] = pt_eq
            elif obj.init is not None:
                # stocks, flows, vars with a defined init
                if isinstance(obj.init, reno.components.Distribution):
                    pt_eq = obj.init.pt(__PTNAME__=f"{obj.qual_name()}_init", **refs)
                else:
                    pt_eq = pm.Deterministic(
                        f"{obj.qual_name()}_init",
                        obj._implied_eq(obj.init).pt(**refs),
                    )
                inits_by_obj[obj] = pt_eq
                refs[obj.qual_name()] = pt_eq
            else:
                # flows and vars without a separate init definition
                # (just run a single normal calculation of their equation)
                pt_eq = pm.Deterministic(
                    f"{obj.qual_name()}_init", obj._implied_eq(obj.eq).pt(**refs)
                )
                inits_by_obj[obj] = pt_eq
                refs[obj.qual_name()] = pt_eq

            # handle initial historical arrays if necessary (but not ones with just -1)
            if obj in historical_ref_taps and len(historical_ref_taps[obj]) > 1:
                # TODO: doesn't handle static but still unclear if that
                # makes sense anyway
                inner_array = [0.0] * (min(historical_ref_taps[obj]) * -1) + [
                    inits_by_obj[obj]
                ]
                hist_var = pm.Deterministic(
                    f"{obj.qual_name()}_hist", pt.stack(inner_array)
                )
                hist_vars_by_obj[obj] = hist_var

        # set up a timestep sequence to pass to scan if any equations are using a TimeRef
        # e.g. t = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        sequences = []
        if model.find_timeref_name() is not None:
            timestep_seq = pt.as_tensor(np.arange(1, timesteps))
            sequences = [timestep_seq]

        # ensure all outputs_info are in correct order
        # NOTE: assumed order of variables whenever they're being passed or returned
        # is order found in model.stocks + model.flows + model.vars
        inits = []
        for obj in model.all_refs():
            if obj in hist_vars_by_obj:
                inits.append(
                    dict(initial=hist_vars_by_obj[obj], taps=historical_ref_taps[obj])
                )
            elif obj in inits_by_obj:
                inits.append(inits_by_obj[obj])

        statics = []  # statics get passed into scan target as non_sequences
        for obj in model.all_flows() + model.all_vars():
            if obj in statics_by_obj:
                statics.append(statics_by_obj[obj])

        seqs, updates = pytensor.scan(
            fn=pt_sim_step(model),
            sequences=sequences,
            non_sequences=statics,
            outputs_info=inits,
            strict=True,
            n_steps=model.steps - 1,
        )

        # handle weird error when only a single stock returned (not wrapping
        # in a list correctly for some reason)
        if len(inits) == 1:
            seqs = [seqs]

        # assign the fully simulated sequences of all non-static refs to deterministics
        # (order is as above, always model.stocks + model.flows + model.vars)
        for i, obj in enumerate(
            model.all_stocks()
            + [
                obj
                for obj in model.all_flows() + model.all_vars()
                if not obj.is_static()
            ]
        ):
            full_seq = pm.Deterministic(
                obj.qual_name(),
                pt.concatenate([pt.as_tensor([inits_by_obj[obj]]), seqs[i].flatten()]),
                dims="t",
            )

            # set pt ref to full sequence for any subsequent metric equations
            refs[obj.qual_name()] = full_seq

        # dependency sort metric equations
        metrics = reno.utils.dependency_compute_order([*model.all_metrics()])
        for metric in metrics:
            metric_var = pm.Deterministic(metric.qual_name(), metric.eq.pt(**refs))
            refs[metric.qual_name()] = metric_var

    # add any likelihoods + observations if relevant
    if observations is not None:
        for observation in observations:
            observation.add_tensors(m)

    return m


def pymc_model_imports() -> str:
    """Get the string of code for the imports needed for ``to_pymc_model_str()`` to
    run correctly."""
    import_strs = [
        "import pytensor",
        "import pytensor.tensor as pt",
        "from pytensor.ifelse import ifelse",
        "import pymc as pm",
        "import numpy as np",
    ]
    return "\n".join(import_strs)


# TODO: wrap in function option (string function would return model)
# TODO: option to add in necessary imports
# TODO: can you run black formatting programmatically on a string?
def to_pymc_model_str(
    model: "reno.model.Model", observations: list[reno.ops.Observation] = None
) -> str:
    """Construct a string of python code to create a pymc model wrapping this system dynamics
    model. Should be a functional (string) equivalent of the ``to_pymc_model()`` function. Includes
    the output from ``pt_sim_step_str()``.

    Expected imports for the resulting code to run:
        >>> import pytensor
        >>> import pytensor.tensor as pt
        >>> from pytensor.ifelse import ifelse
        >>> import pymc as pm
        >>> import numpy as np

    (Alternatively get the string code for these imports with ``pymc_model_imports()``)
    """
    name = "pymc_m" if model.name is None else f"{model.name}_pymc_m"
    step_name = "step" if model.name is None else f"{model.name}_step"
    code = pt_sim_step_str(model) + "\n"
    code += f'coords = {{"t": range({model.steps})}}\n'
    code += f"with pm.Model(coords=coords) as {name}:\n"

    historical_ref_taps = find_historical_tracked_refs(model)

    initial_refs_dict = {}
    if model.find_timeref_name() is not None:
        initial_refs_dict[model.find_timeref_name()] = "pt.as_tensor(0)"

    initial_refs_dict["__PT_SEQ_LEN__"] = f"pt.as_tensor({model.steps})"
    # there are a few operations that need to know the total length
    # of the simulation/sequence in order to create the correct
    # pytensor equivalent output, see ops.py

    ref_compute_order = model.dependency_compute_order(inits_order=True)
    code += "\t# Initial values/timestep 0 equations\n"
    for obj in ref_compute_order:
        if isinstance(obj, reno.components.Stock) and obj.init is None:
            inner_eq = "pt.as_tensor(np.array(0.0))"
            code += f'\t{obj.qual_name()}_init = pm.Deterministic("{obj.qual_name()}_init", {inner_eq})\n'
            initial_refs_dict[obj.qual_name()] = f"{obj.qual_name()}_init"
        elif obj.is_static():
            if isinstance(obj.eq, reno.components.Distribution):
                inner_eq = obj.eq.pt_str(
                    __PTNAME__=obj.qual_name(), **initial_refs_dict
                )
                code += f"\t{obj.qual_name()} = {inner_eq}\n"
            else:
                inner_eq = obj._implied_eq(obj.eq).pt_str(**initial_refs_dict)
                code += f'\t{obj.qual_name()} = pm.Deterministic("{obj.qual_name()}", {inner_eq})\n'
            initial_refs_dict[obj.qual_name()] = obj.qual_name()
        elif obj.init is not None:
            # stocks, flows, vars with a defined init
            if isinstance(obj.init, reno.components.Distribution):
                inner_eq = obj.init.pt_str(
                    __PTNAME__=obj.qual_name() + "_init", **initial_refs_dict
                )
                code += f"\t{obj.qual_name()}_init = {inner_eq}\n"
            else:
                inner_eq = obj._implied_eq(obj.init).pt_str(**initial_refs_dict)
                code += f'\t{obj.qual_name()}_init = pm.Deterministic("{obj.qual_name()}_init", {inner_eq})\n'
            initial_refs_dict[obj.qual_name()] = f"{obj.qual_name()}_init"
        else:
            # flows and vars without a separate init definition
            inner_eq = obj._implied_eq(obj.eq).pt_str(**initial_refs_dict)
            code += f'\t{obj.qual_name()}_init = pm.Deterministic("{obj.qual_name()}_init", {inner_eq})\n'
            initial_refs_dict[obj.qual_name()] = f"{obj.qual_name()}_init"

        # handle initial historical arrays if necessary (but not ones with just -1)
        if obj in historical_ref_taps and len(historical_ref_taps[obj]) > 1:
            # TODO: doesn't handle static but still unclear if that makes sense anyway
            inner_array = (
                ", ".join(["0.0"] * (min(historical_ref_taps[obj]) * -1))
                + f", {obj.qual_name()}_init"
            )
            inner_eq = f"pt.stack([{inner_array}])"
            code += f'\t{obj.qual_name()}_hist = pm.Deterministic("{obj.qual_name()}_hist", {inner_eq})\n'

    # add timeseries
    if model.find_timeref_name() is not None:
        code += f"\ttimestep_seq = pt.as_tensor(np.arange(1, {model.steps}))\n"

    # call the scan function
    return_names = ", ".join(
        [f"{obj.qual_name()}_seq" for obj in model.all_refs() if not obj.is_static()]
    )
    if (
        len(
            model.stocks
            + [
                obj
                for obj in model.all_flows() + model.all_vars()
                if not obj.is_static()
            ]
        )
        > 1
    ):
        return_names = f"[{return_names}]"

    # if we don't have to worry about historical values, the outputs_info is
    # just the array of inits
    if len(historical_ref_taps) == 0:
        init_names = ", ".join(
            [
                f"{obj.qual_name()}_init"
                for obj in model.all_refs()
                if not obj.is_static()
            ]
        )
    else:
        init_names = []
        for obj in model.all_refs():
            # TODO: (2025.02.11) there's some mixed up logic here/not
            # handling if we have a historical value on a static variable
            # (which doesn't truely make sense but also shouldn't be
            # explicitly disallowed?)
            if obj in historical_ref_taps and len(historical_ref_taps[obj]) > 1:
                taps = historical_ref_taps[obj]
                init_names.append(f"dict(initial={obj.qual_name()}_hist, taps={taps})")
            elif not obj.is_static():
                init_names.append(f"{obj.qual_name()}_init")
        init_names = ", ".join(init_names)
    code += "\n\t# Run autoregressive step function/timestep-wise updates to fill sequences\n"
    code += f"\t{return_names}, updates = pytensor.scan(\n"
    code += f"\t\tfn={step_name},\n"
    if model.find_timeref_name() is not None:
        code += "\t\tsequences=[timestep_seq],\n"
    code += f"\t\tnon_sequences=[{', '.join([obj.qual_name() for obj in model.all_flows() + model.all_vars() if reno.utils.is_static(obj)])}],\n"
    code += f"\t\toutputs_info=[{init_names}],\n"
    code += "\t\tstrict=True,\n"
    code += f"\t\tn_steps={model.steps - 1}\n"
    code += "\t)\n"

    code += "\n\t# Collect full sequence data for all stocks/flows/vars into pymc variables\n"
    for obj in model.all_stocks() + [
        obj for obj in model.all_flows() + model.all_vars() if not obj.is_static()
    ]:
        code += f'\t{obj.qual_name()} = pm.Deterministic("{obj.qual_name()}", pt.concatenate([pt.as_tensor([{obj.qual_name()}_init]), {obj.qual_name()}_seq.flatten()]), dims="t")\n'

        # set pt ref to full sequence for any metrics
        initial_refs_dict[obj.qual_name()] = obj.qual_name()

    # dependency sort metric equations
    metrics = reno.utils.dependency_compute_order([*model.metrics])
    for metric in metrics:
        code += f'\t{metric.qual_name()} = pm.Deterministic("{metric.qual_name()}", {metric.eq.pt_str(**initial_refs_dict)})\n'
        initial_refs_dict[metric.qual_name()] = metric.qual_name()

    # add any likelihoods + observations if relevant
    if observations is not None:
        for observation in observations:
            code += f"\t{observation.ref.qual_name()}_likelihood = {observation.pt_str(**initial_refs_dict)}\n"

    return code


def find_historical_tracked_refs(
    model: "reno.model.Model",
) -> dict[reno.components.TrackedReference, list[int]]:
    """Get a dictionary of all tracked references in the model used in historical values,
    and get the associated tap list for each. (Taps is what PyTensor calls indices
    for previous sequence values, so you this is used for when an equation calls for
    some stock value at t-5, etc.)

    This is primarily used for population of pymc scan step and scan arguments."""

    historical_values = []
    for ref in model.all_refs():
        historical_values.extend(ref.find_refs_of_type(reno.components.HistoricalValue))

    ref_taps = {}
    for val in historical_values:
        # for now we only directly support a `t - some_static_amount`
        # NOTE: expanding this would likely require always passing in the full
        # time series, much more annoying boilerplate I suspect.
        if not reno.utils.check_for_easy_static_time_eq(val.index_eq):
            warnings.warn(
                f"Historical value {val.qual_name()} has a non-basic time equation which isn't supported in pymc model conversion yet. Please index with a 't - [static_eq]' equation."
            )
        if val.tracked_ref not in ref_taps:
            ref_taps[val.tracked_ref] = []

        index_offset = val.index_eq.eval(t=0, force=True)
        ref_taps[val.tracked_ref].append(index_offset)

    # remove duplicate taps, ensure -1 is in there (we always want last val
    # of every ref in case it's used in a stock eq)
    ref_taps = {ref: list(set(taps + [-1])) for ref, taps in ref_taps.items()}

    # sort each taps list (order of taps is order they're passed to scan target
    # function, so we should keep consistent sorted ordering for the sanity of everyone)
    for ref, taps in ref_taps.items():
        taps.sort()

    return ref_taps


def expected_pt_scan_arg_names(
    model: "reno.model.Model", as_dict: bool = False, tap_1_only: bool = False
) -> list[str] | dict[str, str]:
    """Pytensor scan target functions are called with a very specific ordering of function
    arguments (see https://pytensor.readthedocs.io/en/latest/library/scan.html).

    Assuming a scan function signature of ``def scan(*args):``, this function
    returns a list of the appropriate reference names to put on the left side of
    a ``refname1, refname2, ... = args`` line.

    Note that with as_dict, it returns a dictionary of reference strings. This is used directly
    by pt_sim_step_str, but also includes refs for tap -1 references (to avoid being updated to
    the "current" timestep) so this is useful for pt_sim_step too

    so with as_dict False:
        [t, v0, v1, v2_h2, v2, ...]
    with as_dict True, tap_1_only False:
        {"t": "t", "v0": "v0", "v1_h1": "v1", "v1": "v1", "v2_h2": "v2_h2", "v2_h1": "v2", "v2": "v2", ...}
    with as_dict True, tap_1_only True:
        {"v1_h1": "v1", "v2_h1": "v2"}
    """

    # remember ordering is sequences/seq taps, output taps, non-sequence args
    names = []
    names_as_dict = {}
    # initially only filled with the mirrored -1 taps, if actually returning
    # as dict, extend with zipped names

    historical_ref_taps = find_historical_tracked_refs(model)

    # -----------------------
    # SEQUENCES/SEQUENCE TAPS
    # -----------------------

    timeref_name = model.find_timeref_name()
    if timeref_name is not None:
        names.append(timeref_name)

    # -----------------------
    # OUTPUT TAPS (non-static references)
    # -----------------------

    # get last of each stock
    for stock in model.all_stocks():
        if stock in historical_ref_taps:
            for tap in historical_ref_taps[stock]:
                str_tap = str(tap * -1)
                name = f"{stock.qual_name()}_h{str_tap}"
                if str_tap == "1":
                    # special handling for -1 tap
                    names_as_dict[name] = stock.qual_name()
                else:
                    names.append(name)
        names.append(stock.qual_name())

    for flow in model.all_flows():
        if not flow.is_static():
            if flow in historical_ref_taps:
                for tap in historical_ref_taps[flow]:
                    str_tap = str(tap * -1)
                    name = f"{flow.qual_name()}_h{str_tap}"
                    if str_tap == "1":
                        # special handling for -1 tap
                        names_as_dict[name] = flow.qual_name()
                    else:
                        names.append(name)
            names.append(flow.qual_name())

    for var in model.all_vars():
        if not var.is_static():
            if var in historical_ref_taps:
                for tap in historical_ref_taps[var]:
                    str_tap = str(tap * -1)
                    name = f"{var.qual_name()}_h{str_tap}"
                    if str_tap == "1":
                        # special handling for -1 tap
                        names_as_dict[name] = var.qual_name()
                    else:
                        names.append(name)
            names.append(var.qual_name())

    # -----------------------
    # NON-SEQUENCE ARGS (static references)
    # -----------------------

    for flow in model.all_flows():
        if flow.is_static():
            names.append(flow.qual_name())

    for var in model.all_vars():
        if var.is_static():
            names.append(var.qual_name())

    # NOTE: assumed order of variables whenever they're being passed or returned
    # is order found in model.stocks + model.flows + model.vars

    if as_dict and not tap_1_only:
        names_as_dict.update(dict(zip(names, names)))
        return names_as_dict
    elif as_dict and tap_1_only:
        return names_as_dict
    return names


def resolve_pt_scan_args(
    model: "reno.model.Model", args: list[pt.TensorVariable]
) -> dict[str, pt.TensorVariable]:
    """Take the set of arguments passed by pytensor to a target scan function,
    and turn it into a dictionary for passing into reno math components' `pt()`
    calls.

    (This is necessary to ensure the returned pytensor symbolic operations
    are referring to the correct tensorvariables passed into scan.)
    """
    # remember ordering is sequences/seq taps, output taps, non-sequence args
    refs = {}

    arg_names = expected_pt_scan_arg_names(model)
    for i, name in enumerate(arg_names):
        refs[name] = args[i]

    # handle mirrored -1 taps (v1_h1 == v1)
    tap1_mirrors = expected_pt_scan_arg_names(model, True, True)
    for hist1_name, target_name in tap1_mirrors.items():
        refs[hist1_name] = refs[target_name]

    return refs



"""Tests for system dynamics functionality."""

import os

import numpy as np
import pytest

from reno.components import Flow, Piecewise, Scalar, Stock, TimeRef, Variable
from reno.model import Model
from reno.pymc import find_historical_tracked_refs


def test_tub_against_insight_maker():
    """The stock/flow outputs for a system where flows reference stocks etc.
    should match what insight maker outputs, to verify that our n+1, n, n-1
    functionality for various eval'd objects is working as expected."""

    t = TimeRef()

    m = Model("tub", steps=7)
    m.water = Stock()
    m.faucet = Flow(
        Piecewise([m.water + 5, Scalar(0)], [t <= Scalar(4), t > Scalar(4)])
    )
    m.drain = Flow(Scalar(2))

    m.water += m.faucet
    m.water -= m.drain

    ds = m()
    assert ds.water.values.tolist()[0] == [0.0, 3.0, 9.0, 21.0, 45.0, 93.0, 91.0]
    assert ds.faucet.values.tolist()[0] == [5.0, 8.0, 14.0, 26.0, 50.0, 0.0, 0.0]


def test_reassigning_eq_to_ref_warns():
    """Very easy to accidentally define a Flow and then assign its equation to the
    reference itself rather than the ref.eq, so make sure this displays a warning."""

    m = Model("tub", steps=7)
    m.water = Stock()
    m.faucet = Flow()
    m.drain = Flow()

    with pytest.warns(RuntimeWarning):
        m.drain = Scalar(2)


def test_multimodel():
    # TODO: see notebook 11
    pass


def test_copy_model_works(tub_model):
    """Getting a deepcopy of a model shouldn't fail."""
    tub_model.copy("new_name")


def test_historical_tracked_refs():
    """A basic model with several index equations for historical values should
    correctly construct a taps dictionary for the tracked references."""
    t = TimeRef()

    m = Model()
    m.v0 = Variable(Scalar(1))
    m.v1 = Variable(t)
    m.v2 = Variable(m.v1.history(t - 1))
    m.v3 = Variable(m.v1.history(t - 2))
    m.v4 = Variable(m.v2.history(t - m.v0))

    assert find_historical_tracked_refs(m) == {m.v1: [-2, -1], m.v2: [-1]}


def test_multiple_consecutive_runs(tub_model):
    """Running a model multiple times should work the same each time/each
    consecutive run starting from scratch"""
    run1 = tub_model(n=10, faucet_volume=6)
    run2 = tub_model(n=10, faucet_volume=6)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_multiple_multimodel_consecutive_runs(tub_multimodel):
    """Running a multimodel multiple times should work the same each time/each
    consecutive run starting from scratch"""
    run1 = tub_multimodel(n=10, faucet_volume=6)
    assert isinstance(tub_multimodel.after_drain.loss_multiplier.value, np.ndarray)
    assert isinstance(tub_multimodel.after_drain.loss.value, np.ndarray)
    run2 = tub_multimodel(n=10, faucet_volume=6)
    assert isinstance(tub_multimodel.after_drain.loss_multiplier.value, np.ndarray)
    assert isinstance(tub_multimodel.after_drain.loss.value, np.ndarray)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_consecutive_scalar_collapse():
    """For unknown reasons, an evaluated scalar variable after a run seems to return a
    raw float instead of a numpy array with a single value."""
    m = Model()
    m.v = Variable(1.0)

    m()
    assert isinstance(m.v.value, np.ndarray)
    m()
    assert isinstance(m.v.value, np.ndarray)


def test_subsequent_pymc_prior_calls(tub_model):
    """Running the same pymc model multiple times should have the same results."""
    run1 = tub_model.pymc(10, compute_prior_only=True)
    run2 = tub_model.pymc(10, compute_prior_only=True)

    for variable in run1.prior.keys():
        assert (run1.prior[variable].values == run2.prior[variable].values).all()


def test_symmetric_config(tub_model):
    """Restoring a previous configuration should _exactly_ (symmetrically) restore
    values. This is most obvious in what gets passed to pymc models."""
    previous = tub_model.config()
    assert tub_model.water_level.init is None
    tub_model.config(**previous)
    assert tub_model.water_level.init is None


def test_symmetric_config_pymc_str(tub_model):
    """Restoring a previous configuration should _exactly_ (symmetrically) restore
    values. This is most obvious in what gets passed to pymc models."""
    previous = tub_model.config()
    str1 = tub_model.pymc_str()
    tub_model.config(**previous)
    str2 = tub_model.pymc_str()

    assert str1 == str2


def test_to_from_dict_consistency(tub_multimodel):
    """The serialized from of a deserialized serialized model should
    be the same as the original serialization."""
    serialized1 = tub_multimodel.to_dict()

    new_model = Model.from_dict(serialized1)
    serialized2 = new_model.to_dict()

    assert serialized1 == serialized2


def test_from_dict_result_consistency(tub_multimodel):
    """Loading a previously saved model should have the same results when run."""

    serialized = tub_multimodel.to_dict()
    new_model = Model.from_dict(serialized)

    run1 = tub_multimodel(n=10, faucet_volume=6)
    run2 = new_model(n=10, faucet_volume=6)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()


def test_save_load(data_file_loc, tub_multimodel):
    """Saving to and loading from a file should have the same results when run."""
    run1 = tub_multimodel(n=10, faucet_volume=6)
    filename = f"{data_file_loc}/testmodel.json"

    tub_multimodel.save(filename)
    assert os.path.exists(filename)

    new_model = Model.load(filename)
    run2 = new_model(n=10, faucet_volume=6)

    for variable in run1.keys():
        assert (run1[variable].values == run2[variable].values).all()



"""Classes for math operations, these are used to build up symbolic equations,
similar in principle to what something like PyTensor is doing."""

import math
import warnings

import numpy as np
import pymc as pm
import pytensor.tensor as pt

import reno

__all__ = [
    # -- series math operations --
    "series_max",
    "series_min",
    "sum",
    "index",
    "slice",
    # -- normal math operations --
    "add",
    "sub",
    "mul",
    "div",
    "mod",
    "abs",
    "lt",
    "lte",
    "gt",
    "gte",
    "eq",
    "ne",
    "bool_and",
    "bool_or",
    "minimum",
    "maximum",
    "clip",
    "log",
    "sin",
    "interpolate",
    "assign",
    # -- distributions --
    "Normal",
    "Uniform",
    "DiscreteUniform",
    "Bernoulli",
    "Categorical",
    "List",
    "Observation",
    # "Sweep",
]


# ==================================================
# SERIES MATH OPERATIONS
# ==================================================
# These are primarily meant for metric equations that run after a
# simulation and operate on a whole timeseries at a time, rather
# than calculating for a single timestep


# NOTE: can't use 'proper' name of max because TrackedReferences already have a
# max (equation max) which I don't want to rename.
class series_max(reno.components.Operation):
    """Maximum value throughout time series. Effectively a row-wise np.max."""

    OP_REPR = "max"

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs):
        return f"\\text{{max}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(**kwargs)
        return np.max(value, axis=1)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.max(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.max({self.sub_equation_parts[0].pt_str(**refs)})"


# NOTE: can't use 'proper' name of min because TrackedReferences already have a
# min (equation min) which I don't want to rename.
class series_min(reno.components.Operation):
    """Minimum value throughout time series. Effectively a row-wise np.min."""

    OP_REPR = "min"

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs):
        return f"\\text{{min}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(**kwargs)
        return np.min(value, axis=1)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.min(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.min({self.sub_equation_parts[0].pt_str(**refs)})"


class sum(reno.components.Operation):
    """Series-wise sum (e.g. row-wise if a matrix)."""

    # NOTE: sums of static values won't give you a multiple by timesteps by default, e.g.
    # if you have a static variable with value 5, the variable.sum() will return 5. In order
    # to expand a static value based on the timeseries, access a slice of the variable. For
    # instance, to get the full series of static values, use variable[:].sum()

    # NOTE: NOTE: nope, since changed to make operand automatically a slice if it's static.
    # Remember that this operation is really only meant for metrics, use history components
    # for equations within the stocks/flows

    def __init__(self, a):
        if a.is_static() and not isinstance(a, reno.ops.slice):
            super().__init__(reno.ops.slice(a))
        else:
            super().__init__(a)

    def latex(self, **kwargs):
        return f"\\Sigma {self.sub_equation_parts[0].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(**kwargs)
        # a 2d input is assumed at this point because of the slice insertion in
        # init for statics
        return np.sum(value, axis=1)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.sum(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.sum({self.sub_equation_parts[0].pt_str(**refs)})"


# TODO: figure out how to make this work with history instead?
class index(reno.components.Operation):
    """Get a previous value in the time series at specified index, only works for tracked references
    inside of equations for metrics."""

    def __init__(self, a, ind):
        super().__init__(a, ind)

    def latex(self, **kwargs) -> str:
        return f"{self.sub_equation_parts[0].latex(**kwargs)}[{self.sub_equation_parts[1].latex(**kwargs)}]"

    def op_eval(self, **kwargs):
        # TODO: support for static?
        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(**kwargs)
        return value[:, self.sub_equation_parts[1].value]  # TODO: eval sub_eq_parts[1]?

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs)[
            self.sub_equation_parts[1].pt(**refs)
        ]

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"{self.sub_equation_parts[0].pt_str(**refs)}[{self.sub_equation_parts[1].pt_str(**refs)}]"


class slice(reno.components.Operation):
    """Get a slice (range from index start to index stop) of values from timeseries.
    NOTE: only works for tracked references inside of equations for metrics."""

    def __init__(self, a, start=None, stop=None):
        self.start = start
        self.stop = stop

        if self.start is not None:
            # keep these ensure_scalars since start/stop stored
            # on this class, and the ensure_scalar applied to sub_equation_parts
            # in EquationPart obv won't magically apply to these
            self.start = reno.utils.ensure_scalar(self.start)

        if self.stop is not None:
            self.stop = reno.utils.ensure_scalar(self.stop)

        operands = [a]
        if start is not None:
            operands.append(self.start)
        if stop is not None:
            operands.append(self.stop)
        super().__init__(*operands)

    def latex(self, **kwargs) -> str:
        start = "" if self.start is None else self.start.latex(**kwargs)
        stop = "" if self.stop is None else self.stop.latex(**kwargs)
        return f"{self.sub_equation_parts[0].latex(**kwargs)}[{start}:{stop}]"

    # TODO: this probably won't work for row_indices?
    def op_eval(self, t, **kwargs):
        t_0 = 0 if self.start is None else self.start.eval(t, **kwargs)
        t_n = t + 1 if self.stop is None else self.stop.eval(t, **kwargs)
        # we need t + 1 because otherwise .eval(0) of a static with no stop
        # would be an empty array

        # when we have to calculate the indices, they result in arrays
        if isinstance(t_0, np.ndarray):
            t_0 = t_0[0]
        if isinstance(t_n, np.ndarray):
            t_n = t_n[0]
        count = t_n - t_0

        value = self.sub_equation_parts[0].value
        if value is None:
            value = self.sub_equation_parts[0].eval(t, **kwargs)

        # if it's static/vector instead of matrix, we just repeat by timesteps
        if isinstance(value, (float, int)):
            return np.array([np.array([value]).repeat(count)])
        if len(value.shape) == 1:
            return value[:, None].repeat(count, axis=1)
        return value[:, t_0:t_n]

    # NOTE: generally dealing with arrays instead of matrices throughout
    # pytensor because backend is dealing with whatever optimizations for sample
    # math, rather than Reno's math system doing batch math on matrices.
    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        if self.sub_equation_parts[0].is_static():
            seq_length = refs["__PT_SEQ_LEN__"]

            t_0 = 0 if self.start is None else self.start.pt(**refs)
            t_n = seq_length if self.stop is None else self.stop.pt(**refs)
            count = t_n - t_0

            return pt.repeat(self.sub_equation_parts[0].pt(**refs), count)

        if self.start is None and self.stop is None:
            return self.sub_equation_parts[0].pt(**refs)
        t_0 = pt.as_tensor(0) if self.start is None else self.start.pt(**refs)
        if self.stop is None:
            return self.sub_equation_parts[0].pt(**refs)[t_0:]
        return self.sub_equation_parts[0].pt(**refs)[t_0 : self.stop.pt(**refs)]

    def pt_str(self, **refs: dict[str, str]) -> str:
        if self.sub_equation_parts[0].is_static():
            seq_length = refs["__PT_SEQ_LEN__"]

            count_str = (
                str(seq_length) if self.stop is None else self.stop.pt_str(**refs)
            )
            if self.start is not None:
                count_str = f"({count_str} - {self.start.pt_str(**refs)})"

            return (
                f"(pt.repeat({self.sub_equation_parts[0].pt_str(**refs)}, {count_str})"
            )
        if self.start is None and self.stop is None:
            return f"{self.sub_equation_parts[0].pt_str(**refs)}"
        t0 = "pt.as_tensor(0)" if self.start is None else self.start.pt_str(**refs)
        if self.stop is None:
            return f"{self.sub_equation_parts[0].pt_str(**refs)}[{t0}:]"
        return f"{self.sub_equation_parts[0].pt_str(**refs)}[{t0}:{self.stop.pt_str(**refs)}]"


# ==================================================
# NORMAL MATH OPERATIONS
# ==================================================


class add(reno.components.Operation):
    """a + b"""

    OP_REPR = "+"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} + {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) + self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) + self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} + {self.sub_equation_parts[1].pt_str(**refs)})"


class sub(reno.components.Operation):
    """a - b"""

    OP_REPR = "-"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} - {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) - self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) - self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} - {self.sub_equation_parts[1].pt_str(**refs)})"


class mul(reno.components.Operation):
    """a * b"""

    OP_REPR = "*"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} * {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) * self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) * self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} * {self.sub_equation_parts[1].pt_str(**refs)})"


class div(reno.components.Operation):
    """a / b"""

    OP_REPR = "/"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"\\frac{{{self.sub_equation_parts[0].latex(**kwargs)}}}{{{self.sub_equation_parts[1].latex(**kwargs)}}}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) / self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) / self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} / {self.sub_equation_parts[1].pt_str(**refs)})"


class mod(reno.components.Operation):
    """a % b"""

    OP_REPR = "%"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\% {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) % self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.mod(
            self.sub_equation_parts[0].pt(**refs), self.sub_equation_parts[1].pt(**refs)
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.mod({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class abs(reno.components.Operation):
    """|a| (absolute value)"""

    def __init__(self, a):
        super().__init__("abs", a)

    def latex(self, **kwargs):
        return f"|{self.sub_equation_parts[0].latex(**kwargs)}|"

    def op_eval(self, **kwargs):
        return np.abs(self.sub_equation_parts[0].eval(**kwargs))

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.abs(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.abs({self.sub_equation_parts[0].pt_str(**refs)})"


class lt(reno.components.Operation):
    """a < b"""

    OP_REPR = "<"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} < {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) < self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) < self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} < {self.sub_equation_parts[1].pt_str(**refs)})"


class lte(reno.components.Operation):
    """a <= b"""

    OP_REPR = "<="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\leq {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) <= self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) <= self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} <= {self.sub_equation_parts[1].pt_str(**refs)})"


class gt(reno.components.Operation):
    """a > b"""

    OP_REPR = ">"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} > {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) > self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) > self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} > {self.sub_equation_parts[1].pt_str(**refs)})"


class gte(reno.components.Operation):
    """a >= b"""

    OP_REPR = ">="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\geq {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) >= self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) >= self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} >= {self.sub_equation_parts[1].pt_str(**refs)})"


class eq(reno.components.Operation):
    """a == b"""

    OP_REPR = "=="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} = {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) == self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.eq(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.eq({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class ne(reno.components.Operation):
    """a != b"""

    OP_REPR = "!="

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\neq {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) != self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.neq(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.ne({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


# TODO: rename to just and?
class bool_and(reno.components.Operation):
    """a and b"""

    OP_REPR = "and"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\text{{ and }} {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) & self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) & self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} & {self.sub_equation_parts[1].pt_str(**refs)})"


# TODO: rename to just or?
class bool_or(reno.components.Operation):
    """a or b"""

    OP_REPR = "or"

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"{self.sub_equation_parts[0].latex(**kwargs)} \\text{{ or }} {self.sub_equation_parts[1].latex(**kwargs)}"

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs) | self.sub_equation_parts[
            1
        ].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs) | self.sub_equation_parts[1].pt(
            **refs
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"({self.sub_equation_parts[0].pt_str(**refs)} | {self.sub_equation_parts[1].pt_str(**refs)})"


class minimum(reno.components.Operation):
    """Element-wise minimum of array elements between two arrays or values, same as np.minimum."""

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"\\text{{min}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.minimum(
            self.sub_equation_parts[0].eval(**kwargs),
            self.sub_equation_parts[1].eval(**kwargs),
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.minimum(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.minimum({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class maximum(reno.components.Operation):
    """Element-wise maximum of array elements between two arrays or values, same as np.maximum."""

    def __init__(self, a, b):
        super().__init__(a, b)

    def latex(self, **kwargs):
        return f"\\text{{maximum}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.maximum(
            self.sub_equation_parts[0].eval(**kwargs),
            self.sub_equation_parts[1].eval(**kwargs),
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.maximum(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.maximum({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)})"


class clip(reno.components.Operation):
    """Simultaneously apply upper and lower bound constraint (element-wise)."""

    def __init__(self, a, b, c):
        super().__init__(a, b, c)

    def latex(self, **kwargs):
        return f"\\text{{clip}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}, {self.sub_equation_parts[2].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.clip(
            self.sub_equation_parts[0].eval(**kwargs),
            self.sub_equation_parts[1].eval(**kwargs),
            self.sub_equation_parts[2].eval(**kwargs),
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.clip(
            self.sub_equation_parts[0].pt(**refs),
            self.sub_equation_parts[1].pt(**refs),
            self.sub_equation_parts[2].pt(**refs),
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.clip({self.sub_equation_parts[0].pt_str(**refs)}, {self.sub_equation_parts[1].pt_str(**refs)}, {self.sub_equation_parts[2].pt_str(**refs)})"


class log(reno.components.Operation):
    """ln(a) (natural log, naming it log because this is pytensor's and numpy's default)"""

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs):
        return f"\\text{{ln}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.log(self.sub_equation_parts[0].eval(**kwargs))

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.log(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.log({self.sub_equation_parts[0].pt_str(**refs)})"


class sin(reno.components.Operation):
    """sin(a)"""

    def __init__(self, a):
        super().__init__(reno.utils.ensure_scalar(a))

    def latex(self, **kwargs):
        return f"\\text{{sin}}({self.sub_equation_parts[0].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.sin(self.sub_equation_parts[0].eval(**kwargs))

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.sin(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.sin({self.sub_equation_parts[0].pt_str(**refs)})"


# TODO: all the other trig functions


class interpolate(reno.components.Operation):
    """Given a dataset of x -> y datapoints, interpolate any new data along the line formed by the points.
    Equivalent to numpy's interp function.

    Args:
        x: The input x-coordinates that you want interpolated into y outputs.
        x_data: The x-coordinate data to base interpolation on.
        y_data: the y-coordinate data to base interpolation on.
    """

    def __init__(self, x, x_data: list | np.ndarray, y_data: list | np.ndarray):
        super().__init__(x, x_data, y_data)

    def latex(self, **kwargs):
        return f"\\text{{interpolate}}({self.sub_equation_parts[0].latex(**kwargs)}, {self.sub_equation_parts[1].latex(**kwargs)}, {self.sub_equation_parts[2].latex(**kwargs)})"

    def op_eval(self, **kwargs):
        return np.interp(
            self.sub_equation_parts[0].eval(**kwargs),
            self.sub_equation_parts[1].eval(**kwargs),
            self.sub_equation_parts[2].eval(**kwargs),
        )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pt.interpolate1d(
            self.sub_equation_parts[1].pt(**refs),
            self.sub_equation_parts[2].pt(**refs),
            extrapolate=False,
        )(self.sub_equation_parts[0].pt(**refs))

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f"pt.interpolate1d({self.sub_equation_parts[1].pt_str(**refs)}, {self.sub_equation_parts[2].pt_str(**refs)},)({self.sub_equation_parts[0].pt_str(**refs)})"


class assign(reno.components.Operation):
    """This is to handle the weird seek_refs issues when you just set a tracked ref's
    equation to another tracked ref. By "wrapping" it in an effectively blank operation,
    this mitigates the annoying recursion issue."""

    OP_REPR = "="

    def __init__(self, a):
        super().__init__(a)

    def latex(self, **kwargs):
        return self.sub_equation_parts[0].latex(**kwargs)

    def op_eval(self, **kwargs):
        return self.sub_equation_parts[0].eval(**kwargs)

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return self.sub_equation_parts[0].pt(**refs)

    def pt_str(self, **refs: dict[str, str]) -> str:
        return self.sub_equation_parts[0].pt_str(**refs)


# ==================================================
# DISTRIBUTIONS
# ==================================================


class Normal(reno.components.Distribution):
    def __init__(self, mean, std=1.0):
        super().__init__()
        self.mean = mean
        self.std = std

    def latex(self, **kwargs):
        return f"\\mathcal{{N}}({self.mean}, {self.std}^2)"

    def populate(self, n):
        self.value = np.random.normal(self.mean, self.std, size=(n,))

    def __repr__(self):
        return f"Normal({self.mean}, {self.std})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        return pm.Normal(name, self.mean, self.std)

    def pt_str(self, **refs: dict[str, str]) -> str:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        return f'pm.Normal("{name}", {self.mean}, {self.std})'


class Uniform(reno.components.Distribution):
    def __init__(self, low=0.0, high=1.0):
        super().__init__()
        self.low = low
        self.high = high

    def latex(self, **kwargs):
        return f"\\mathcal{{U}}({self.low}, {self.high})"

    def populate(self, n):
        self.value = np.random.uniform(self.low, self.high, size=(n,))

    def __repr__(self):
        return f"Uniform({self.low}, {self.high})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        return pm.Uniform(name, self.low, self.high)

    def pt_str(self, **refs: dict[str, str]) -> str:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        return f'pm.Uniform("{name}", {self.low}, {self.high})'


class DiscreteUniform(reno.components.Distribution):
    """Low is inclusive, high is exclusive."""

    def __init__(self, low: int = 0, high: int = 2):
        super().__init__()
        self.low = low
        self.high = high

    def latex(self, **kwargs):
        return f"\\text{{DiscreteUniform}}({self.low}, {self.high})"

    def populate(self, n: int):
        self.value = np.random.randint(self.low, self.high, n)

    def __repr__(self):
        return f"DiscreteUniform({self.low}, {self.high})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        return pm.DiscreteUniform(name, self.low, self.high)

    def pt_str(self, **refs: dict[str, str]) -> str:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        return f'pm.DiscreteUniform("{name}", {self.low}, {self.high})'


class Bernoulli(reno.components.Distribution):
    """Discrete single event probability (p is probability of eval == 1)"""

    def __init__(self, p: float, use_p_dist: bool = False):
        super().__init__()
        self.p = p
        self.use_p_dist = use_p_dist

    def latex(self, **kwargs):
        return f"\\text{{Bernoulli}}({self.p})"

    def populate(self, n):
        self.value = np.random.binomial(1, self.p, n)

    def __repr__(self):
        return f"Bernoulli({self.p})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        inner_dist = self.p
        if self.use_p_dist:
            inner_dist = pm.Interpolated(
                f"{name}_p",
                x_points=np.array([0.0, 1.0]),
                pdf_points=np.array([1 - self.p, self.p]),
            )
        return pm.Bernoulli(name, inner_dist)

    def pt_str(self, **refs: dict[str, str]) -> str:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        inner_dist = self.p
        if self.use_p_dist:
            inner_dist = f'pm.Interpolated("{name}_p", x_points=np.array([0.0, 1.0]), pdf_points=np.array([{1 - self.p}, {self.p}]))'
        return f'pm.Bernoulli("{name}", {inner_dist})'


class Categorical(reno.components.Distribution):
    def __init__(self, p: list[float], use_p_dist: bool = False):
        super().__init__()
        self.p = p
        self.use_p_dist = use_p_dist

    def latex(self, **kwargs):
        return f"\\text{{Categorical}}({self.p})"

    def populate(self, n):
        self.value = np.argmax(np.random.multinomial(1, self.p, n), axis=1)

    def __repr__(self):
        return f"Categorical({self.p})"

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        inner_dist = self.p
        if self.use_p_dist:
            inner_dist = pm.Dirichlet(f"{name}_p", self.p)
        return pm.Categorical(name, inner_dist)

    def pt_str(self, **refs: dict[str, str]) -> str:
        name = "dist" + str(id(self))
        if "__PTNAME__" in refs:
            name = refs["__PTNAME__"]
        inner_dist = self.p
        if self.use_p_dist:
            inner_dist = f'pm.Dirichlet("{name}_p", {self.p})'
        return f'pm.Categorical("{name}", {inner_dist})'


class List(reno.components.Distribution):
    """Tile passed list to the sample size so each value is hit roughly
    equally (dependent on exact sample size) and deterministically"""

    def __init__(self, values: list | np.ndarray | set):
        super().__init__()
        self.values = values

    def latex(self, **kwargs):
        return f"{self.values}"

    def populate(self, n):
        repetitions = n / len(self.values)
        # if the specified value is _larger_ than the samples, we have to
        # truncate (and warn, does the user know this is what's happening?)
        if repetitions < 1:
            warnings.warn(
                f"Not enough samples in simulation to hit every value in list '{self.values}', would require at least `n={len(self.values)}`",
                RuntimeWarning,
            )
            expanded = np.array(self.values)
        else:
            expanded = np.tile(self.values, math.ceil(repetitions))
        self.value = expanded[:n]

    def __repr__(self):
        return f"List({self.values})"


class Observation(reno.components.Distribution):
    """Represents a Normal distribution around an observed value.

    Should only be used for supplying observational data with likelihoods
    to bayesian models constructed with model.pymc()

    Args:
        ref (reno.components.Reference): The equation to supply an observed value for.
        sigma (float): The std dev to use for the likelihood Normal distribution.
        data (list): The actual observed data to apply.
    """

    def __init__(
        self, ref: reno.components.Reference, sigma: float = 1.0, data: list = None
    ):
        super().__init__()
        self.ref = ref
        self.sigma = sigma
        self.data = data

    def add_tensors(self, pymc_model):
        with pymc_model:
            # sigma = pm.HalfNormal(f"{self.ref.qual_name()}_sigma", self.sigma)
            pm.Normal(
                f"{self.ref.qual_name()}_likelihood",
                pymc_model[self.ref.qual_name()],
                self.sigma,
                observed=self.data,
            )

    def pt(self, **refs: dict[str, pt.TensorVariable]) -> pt.TensorVariable:
        return pm.Normal(
            f"{self.ref.qual_name()}_likelihood",
            self.ref.pt(**refs),
            self.sigma,
            observed=self.data,
        )

    def pt_str(self, **refs: dict[str, str]) -> str:
        return f'pm.Normal("{self.ref.qual_name()}_likelihood", {self.ref.pt_str(**refs)}, {self.sigma}, observed={self.data})'


# class Sweep(reno.components.Distribution):
#     """Similar in principle to ops.List, unimplemented idea for this is to
#     grab all variables that are sweeps and collectively make sure they are fully
#     permutated, rather than requiring user to manually make sure they change at
#     varying rates between samples."""
#
#     # TODO: semantically tells to ensure that for _all_ sweep dists
#     # in a system, make sure all value combinations are hit (if size allows, warn if
#     # doesn't)
#     def __init__(self, sweep_values: list | np.ndarray | set):
#         super().__init__()
#         self.sweep_values = sweep_values
#
#     def latex(self, **kwargs):
#         return f"\\text{{Sweep}}({self.sweep_values})"
#
#     def populate(self, n):
#         # TODO: this will be difficult to implement, need to find all other
#         # sweeps in system? We have no ref to that
#         pass
#
#     def __repr__(self):
#         return f"Sweep({self.sweep_values})"
#
